import tempfile
from pathlib import Path
import unittest

import numpy as np
import pandas as pd
import scipy.sparse as sp
import anndata as ad
import torch

from stereosegger.cli.convert_saw_h5ad_to_segger_parquet import convert_saw_h5ad_to_parquet
from stereosegger.data.tx_graph import build_grid_gene_bin_edge_index
from stereosegger.data.parquet.sample import STSampleParquet


class TestSawBin1(unittest.TestCase):
    def _make_adata(self) -> ad.AnnData:
        coords = np.array(
            [
                [0, 0],
                [1, 0],
                [0, 1],
                [1, 1],
                [0, 2],
                [1, 2],
            ],
            dtype=float,
        )
        X = sp.csr_matrix(
            [
                [1, 0, 0],
                [2, 0, 0],
                [0, 3, 0],
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
            ],
            dtype=int,
        )
        adata = ad.AnnData(X=X)
        adata.obsm["spatial"] = coords
        adata.var_names = ["g0", "g1", "g2"]
        return adata

    def test_convert_and_grid_edges(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            h5ad_path = tmpdir / "saw.h5ad"
            out_dir = tmpdir / "out"

            adata = self._make_adata()
            adata.write_h5ad(h5ad_path)

            convert_saw_h5ad_to_parquet(
                h5ad_path=h5ad_path,
                out_dir=out_dir,
                bin_pitch=1.0,
                min_count=1,
            )

            transcripts = pd.read_parquet(out_dir / "transcripts.parquet")
            required_cols = {"transcript_id", "x", "y", "bx", "by", "gene_id", "count"}
            self.assertTrue(required_cols.issubset(transcripts.columns))

            bx = transcripts["bx"].to_numpy()
            by = transcripts["by"].to_numpy()
            # Test build_grid_gene_bin_edge_index (star topology)
            edge_index = build_grid_gene_bin_edge_index(bx, by, connectivity=4, within_bin_edges="star")
            edges = set(map(tuple, edge_index.T.cpu().numpy()))

            # Check if star edges are present for a bin with multiple genes
            # In our data, (0,0) has g0. (1,0) has g0. (0,1) has g1. (1,1) has g2.
            # (0,2) has g0. (1,2) has g1.
            # Wait, my mock data:
            # bin (0,0): g0
            # bin (1,0): g0
            # bin (0,1): g1
            # bin (1,1): g2
            # bin (0,2): g0
            # bin (1,2): g1
            # None of the bins have multiple genes in this mock?
            # Let me check X:
            # row 0: [1, 0, 0] -> g0
            # row 1: [2, 0, 0] -> g0
            # Row 0 and 1 have same coordinates (0,0) and (1,0)? No.
            # coords[0] = (0,0), coords[1] = (1,0)
            
            # Let me adjust mock data to have a bin with 2 genes
            # row 0: (0,0), g0
            # row 1: (0,0), g1 (change this)
            
            adata.X[1, 0] = 0
            adata.X[1, 1] = 5 # (1,0) now has g1
            # Wait, still different bins.
            
            # (0,0) -> g0, g1
            adata.obsm["spatial"][1] = [0, 0]
            adata.write_h5ad(h5ad_path)
            convert_saw_h5ad_to_parquet(h5ad_path=h5ad_path, out_dir=out_dir)
            transcripts = pd.read_parquet(out_dir / "transcripts.parquet")
            
            bx = transcripts["bx"].to_numpy()
            by = transcripts["by"].to_numpy()
            edge_index = build_grid_gene_bin_edge_index(bx, by, connectivity=4, within_bin_edges="star")
            edges = set(map(tuple, edge_index.T.cpu().numpy()))

            # Indices for (0,0) bin
            bin_00_indices = transcripts.index[(transcripts["bx"] == 0) & (transcripts["by"] == 0)].tolist()
            if len(bin_00_indices) >= 2:
                i, j = bin_00_indices[0], bin_00_indices[1]
                self.assertIn((i, j), edges)
                self.assertIn((j, i), edges)

    def test_create_dataset_fast_tile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            h5ad_path = tmpdir / "saw.h5ad"
            base_dir = tmpdir / "base"
            data_dir = tmpdir / "dataset"

            adata = self._make_adata()
            adata.write_h5ad(h5ad_path)

            convert_saw_h5ad_to_parquet(
                h5ad_path=h5ad_path,
                out_dir=base_dir,
                bin_pitch=1.0,
                min_count=1,
            )

            sample = STSampleParquet(
                base_dir=base_dir, 
                sample_type="saw_bin1", 
                n_workers=1,
                allow_missing_boundaries=True
            )
            sample.save(
                data_dir=data_dir,
                tile_width=100,
                tile_height=100,
                tx_graph_mode="grid_bins",
                grid_connectivity=4,
                within_bin_edges="star",
                allow_missing_boundaries=True,
            )

            processed = list((data_dir / "train_tiles" / "processed").glob("*.pt"))
            processed += list((data_dir / "test_tiles" / "processed").glob("*.pt"))
            processed += list((data_dir / "val_tiles" / "processed").glob("*.pt"))
            self.assertTrue(len(processed) > 0)


if __name__ == "__main__":
    unittest.main()


import click
import os
import scanpy as sc
from stereosegger.data.utils import calculate_gene_celltype_abundance_embedding
from stereosegger.cli.utils import add_options, CustomFormatter
from pathlib import Path
import logging
from argparse import Namespace
from stereosegger.data.parquet.sample import STSampleParquet
from typing import Optional
import time

# Path to default YAML configuration file
data_yml = Path(__file__).parent / "configs" / "create_dataset" / "default_fast.yaml"

# CLI command to create a Segger dataset
help_msg = "Create Segger dataset from spatial transcriptomics data (Xenium or MERSCOPE)"


@click.command(name="create_dataset", help=help_msg)
@add_options(config_path=data_yml)
@click.option("--base_dir", type=Path, required=True, help="Directory containing the raw dataset.")
@click.option("--data_dir", type=Path, required=True, help="Directory to save the processed Segger dataset.")
@click.option(
    "--sample_type", type=str, default=None, help='The sample type of the raw data, e.g., "xenium" or "merscope".'
)
@click.option("--scrnaseq_file", type=Path, default=None, help="Path to the scRNAseq file.")
@click.option(
    "--celltype_column", type=str, default=None, help="Column name for cell type annotations in the scRNAseq file."
)
@click.option("--k_bd", type=int, default=3, help="Number of nearest neighbors for boundary nodes.")
@click.option("--dist_bd", type=float, default=15.0, help="Maximum distance for boundary neighbors.")
@click.option("--k_tx", type=int, default=3, help="Number of nearest neighbors for transcript nodes.")
@click.option("--dist_tx", type=float, default=5.0, help="Maximum distance for transcript neighbors.")
@click.option(
    "--tile_size",
    type=int,
    default=None,
    help="If provided, specifies the size of the tile. Overrides `tile_width` and `tile_height`.",
)
@click.option(
    "--tile_width", type=int, default=None, help="Width of the tiles in pixels. Ignored if `tile_size` is provided."
)
@click.option(
    "--tile_height", type=int, default=None, help="Height of the tiles in pixels. Ignored if `tile_size` is provided."
)
@click.option("--neg_sampling_ratio", type=float, default=5.0, help="Ratio of negative samples.")
@click.option("--frac", type=float, default=1.0, help="Fraction of the dataset to process.")
@click.option("--val_prob", type=float, default=0.1, help="Proportion of data for use for validation split.")
@click.option("--test_prob", type=float, default=0.2, help="Proportion of data for use for test split.")
@click.option("--n_workers", type=int, default=1, help="Number of workers for parallel processing.")
@click.option(
    "--tx_graph_mode",
    type=click.Choice(["kdtree", "grid_bins"], case_sensitive=False),
    default="kdtree",
    help="Strategy for transcript-transcript edges.",
)
@click.option(
    "--grid_connectivity",
    type=int,
    default=8,
    help="Grid connectivity (4 or 8) for grid-based transcript graphs.",
)
@click.option(
    "--within_bin_edges",
    type=click.Choice(["none", "star"], case_sensitive=False),
    default="none",
    help="Within-bin edge strategy for grid graphs.",
)
@click.option("--bin_pitch", type=float, default=1.0, help="Bin pitch for grid graph fallbacks.")
@click.option(
    "--allow_missing_boundaries",
    is_flag=True,
    default=False,
    help="Allow dataset creation without boundaries (prediction-only mode).",
)
def create_dataset(args: Namespace):

    # Setup logging
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(CustomFormatter())
    logging.basicConfig(level=logging.INFO, handlers=[ch])

    # If scRNAseq file is provided, calculate gene-celltype embeddings
    gene_celltype_abundance_embedding = None
    if args.scrnaseq_file:
        logging.info("Calculating gene and celltype embeddings...")
        scRNAseq = sc.read(args.scrnaseq_file)
        sc.pp.subsample(scRNAseq, 0.1)
        gene_celltype_abundance_embedding = calculate_gene_celltype_abundance_embedding(scRNAseq, args.celltype_column)

    # Initialize the sample class
    logging.info("Initializing sample...")
    sample = STSampleParquet(
        base_dir=args.base_dir,
        n_workers=args.n_workers,
        sample_type=args.sample_type,
        weights=gene_celltype_abundance_embedding,
        allow_missing_boundaries=args.allow_missing_boundaries,
    )

    # Save Segger dataset
    logging.info("Saving dataset for Segger...")
    start_time = time.time()
    sample.save(
        data_dir=args.data_dir,
        k_bd=args.k_bd,
        dist_bd=args.dist_bd,
        k_tx=args.k_tx,
        dist_tx=args.dist_tx,
        tx_graph_mode=args.tx_graph_mode,
        grid_connectivity=args.grid_connectivity,
        within_bin_edges=args.within_bin_edges,
        bin_pitch=args.bin_pitch,
        allow_missing_boundaries=args.allow_missing_boundaries,
        tile_size=args.tile_size,
        tile_width=args.tile_width,
        tile_height=args.tile_height,
        neg_sampling_ratio=args.neg_sampling_ratio,
        frac=args.frac,
        val_prob=args.val_prob,
        test_prob=args.test_prob,
    )
    end_time = time.time()
    logging.info(f"Time to save dataset: {end_time - start_time} seconds")
    logging.info("Dataset saved successfully.")


if __name__ == "__main__":
    create_dataset()



import click
import numpy as np
import pandas as pd
import scipy.sparse as sp
from pathlib import Path
import scanpy as sc
import geopandas as gpd
from shapely.geometry import shape, Point


def _read_tif(path: Path) -> np.ndarray:
    try:
        import tifffile

        return tifffile.imread(path)
    except ImportError:
        try:
            import imageio.v3 as iio

            return iio.imread(path)
        except ImportError as exc:
            msg = "Reading TIFF requires `tifffile` or `imageio`."
            raise ImportError(msg) from exc


def _vectorize_labels(labels: np.ndarray) -> gpd.GeoDataFrame:
    try:
        import rasterio.features
    except ImportError:
        raise ImportError("Vectorizing labels requires `rasterio`.") from None

    records = []
    # Use cast to int32 for rasterio
    for geom, value in rasterio.features.shapes(labels.astype(np.int32), mask=labels > 0):
        if value == 0:
            continue
        records.append({"boundary_id": int(value), "geometry": shape(geom)})
    gdf = gpd.GeoDataFrame(records, geometry="geometry")
    return gdf


def convert_saw_h5ad_to_parquet(
    h5ad_path: Path,
    out_dir: Path,
    bin_pitch: float = 1.0,
    min_count: int = 1,
    labels_tif: Path | None = None,
    tissue_mask_tif: Path | None = None,
    bbox: tuple[float, float, float, float] | None = None,
    gene_name_source: str = "real_gene_name",
    top_genes: int | None = None,
    max_nnz: int | None = None,
) -> None:
    out_dir = Path(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    adata = sc.read_h5ad(h5ad_path)

    if "spatial" in adata.obsm:
        coords = np.asarray(adata.obsm["spatial"])
    elif {"x", "y"}.issubset(adata.obs.columns):
        coords = adata.obs[["x", "y"]].to_numpy()
    else:
        raise ValueError("Coordinates not found. Expected adata.obsm['spatial'] or adata.obs['x','y'].")

    if bbox is not None:
        xmin, xmax, ymin, ymax = bbox
        mask = (coords[:, 0] >= xmin) & (coords[:, 0] <= xmax) & (coords[:, 1] >= ymin) & (coords[:, 1] <= ymax)
        adata = adata[mask].copy()
        coords = coords[mask]

    if tissue_mask_tif is not None:
        mask_img = _read_tif(tissue_mask_tif)
        coords_round = np.rint(coords).astype(int)
        valid = (
            (coords_round[:, 0] >= 0)
            & (coords_round[:, 0] < mask_img.shape[1])
            & (coords_round[:, 1] >= 0)
            & (coords_round[:, 1] < mask_img.shape[0])
        )
        mask_vals = np.zeros(coords_round.shape[0], dtype=bool)
        mask_vals[valid] = mask_img[coords_round[valid, 1], coords_round[valid, 0]] > 0
        adata = adata[mask_vals].copy()
        coords = coords[mask_vals]

    if top_genes is not None:
        gene_sums = np.asarray(adata.X.sum(axis=0)).ravel()
        top_idx = np.argsort(gene_sums)[::-1][:top_genes]
        adata = adata[:, top_idx].copy()

    # Get gene names
    if gene_name_source in adata.var:
        gene_names = adata.var[gene_name_source].astype(str).values
    else:
        gene_names = adata.var_names.astype(str).values

    X = adata.X
    if sp.issparse(X):
        X = X.tocoo()
    else:
        X = sp.coo_matrix(X)

    keep = X.data >= min_count
    rows = X.row[keep]
    cols = X.col[keep]
    vals = X.data[keep]

    if max_nnz is not None and len(vals) > max_nnz:
        rows = rows[:max_nnz]
        cols = cols[:max_nnz]
        vals = vals[:max_nnz]

    x = coords[rows, 0]
    y = coords[rows, 1]
    bx = np.rint(x / bin_pitch).astype(np.int32)
    by = np.rint(y / bin_pitch).astype(np.int32)

    transcripts_df = pd.DataFrame(
        {
            "transcript_id": np.arange(len(vals), dtype=np.int64),
            "x": x.astype(np.float32),
            "y": y.astype(np.float32),
            "bx": bx,
            "by": by,
            "gene_id": cols.astype(np.int32),
            "count": vals.astype(np.int32),
        }
    )

    # Process boundaries and label transcripts
    if labels_tif is not None:
        print("Vectorizing labels and assigning transcripts...")
        labels = _read_tif(labels_tif)
        boundaries_gdf = _vectorize_labels(labels)
        boundaries_gdf.to_parquet(out_dir / "boundaries.parquet", index=False)
        
        # Spatial Join
        # Create points for transcripts
        # We process in chunks if needed? No, geopandas should handle it or dask-geopandas.
        # But here we use standard geopandas.
        
        # Optimized lookup using labels TIFF directly (much faster than spatial join)
        labels = _read_tif(labels_tif)
        x_idx = np.rint(transcripts_df.x.values).astype(int)
        y_idx = np.rint(transcripts_df.y.values).astype(int)
        
        # Ensure indices are within image bounds
        valid = (x_idx >= 0) & (x_idx < labels.shape[1]) & \
                (y_idx >= 0) & (y_idx < labels.shape[0])
        
        assigned_labels = np.zeros(len(transcripts_df), dtype=int)
        assigned_labels[valid] = labels[y_idx[valid], x_idx[valid]]
        
        transcripts_df["overlaps_nucleus"] = (assigned_labels > 0).astype(int)
        transcripts_df["cell_id"] = np.where(assigned_labels > 0, assigned_labels, -1)
        
    else:
        # If no labels, we can't train supervised.
        # Add dummy columns if needed? Or Segger handles missing?
        # Segger checks 'can_build_labels'. If false, no edge_label_index.
        # If we want to run prediction only, this is fine.
        # If we want to train, we need labels.
        pass

    transcripts_df.to_parquet(out_dir / "transcripts.parquet", index=False)

    genes = pd.DataFrame(
        {"gene_id": np.arange(len(gene_names), dtype=np.int32), "gene_name": gene_names.astype(str)}
    )
    genes.to_parquet(out_dir / "genes.parquet", index=False)


@click.command(name="convert_saw_h5ad_to_segger_parquet")
@click.option("--h5ad", "h5ad_path", type=Path, required=True, help="Path to SAW bin1 h5ad file.")
@click.option("--out_dir", type=Path, required=True, help="Output directory for Segger parquet files.")
@click.option("--bin_pitch", type=float, default=1.0, help="Bin pitch for rounding to grid coordinates.")
@click.option("--min_count", type=int, default=1, help="Minimum count to keep a bin-gene entry.")
@click.option("--labels_tif", type=Path, default=None, help="Optional label TIFF for boundary polygons.")
@click.option("--tissue_mask_tif", type=Path, default=None, help="Optional tissue mask TIFF.")
@click.option("--bbox", type=float, nargs=4, default=None, help="Bounding box xmin xmax ymin ymax.")
@click.option("--gene_name_source", type=str, default="real_gene_name", help="Column in adata.var for gene names.")
@click.option("--top_genes", type=int, default=None, help="Keep only top K genes by total counts.")
@click.option("--max_nnz", type=int, default=None, help="Debug: cap number of non-zero entries.")
def main(
    h5ad_path: Path,
    out_dir: Path,
    bin_pitch: float,
    min_count: int,
    labels_tif: Path | None,
    tissue_mask_tif: Path | None,
    bbox: tuple[float, float, float, float] | None,
    gene_name_source: str,
    top_genes: int | None,
    max_nnz: int | None,
):
    convert_saw_h5ad_to_parquet(
        h5ad_path=h5ad_path,
        out_dir=out_dir,
        bin_pitch=bin_pitch,
        min_count=min_count,
        labels_tif=labels_tif,
        tissue_mask_tif=tissue_mask_tif,
        bbox=bbox,
        gene_name_source=gene_name_source,
        top_genes=top_genes,
        max_nnz=max_nnz,
    )


if __name__ == "__main__":
    main()


from stereosegger.cli.create_dataset_fast import create_dataset
from stereosegger.cli.train_model import train
from stereosegger.cli.predict import predict
import click


# Setup main CLI command
@click.group(help="Command line interface for the StereoSegger segmentation package")
def stereosegger():
    pass


# Add sub-commands to main CLI commands
stereosegger.add_command(create_dataset)
stereosegger.add_command(train)
stereosegger.add_command(predict)



import importlib.util
import dask

# FIX: Disable dask-expr query planning to support spatialdata/squidpy
# Must be set before importing any module that uses dask.dataframe
dask.config.set({'dataframe.query-planning': False})

__version__ = "0.1.0"

cupy_available = importlib.util.find_spec("cupy") is not None

from stereosegger.data import *
from stereosegger.models import *
from stereosegger.training import *
# from stereosegger.validation import *

# segger.prediction requires cupy, which is not available in macOS
if cupy_available:
    from stereosegger.prediction import *

__all__ = ["data", "models", "prediction", "training"]


from scipy.spatial import Rectangle
import shapely
import numpy as np
import math


class NDTree:
    """
    NDTree is a data structure for recursively splitting multi-dimensional data
    into smaller regions until each leaf node contains less than or equal to a
    specified number of points. It stores these regions in a balanced binary
    tree.

    Attributes
    ----------
    data : np.ndarray
        The input data to be partitioned.
    n : int
        The maximum number of points allowed in a leaf node.
    idx : np.ndarray
        The indices of the input data points.
    boxes : list
        A list to store the bounding boxes (as shapely polygons) of each region
        in the tree.
    rect : Rectangle
        The bounding box of the entire input data space.
    tree : innernode
        The root of the NDTree.
    """

    def __init__(self, data, n):
        """
        Initializes the NDTree with the given data and maximum points per leaf
        node.

        Parameters
        ----------
        data : np.ndarray
            The input data to be partitioned.
        n : int
            The maximum number of points allowed in a leaf node.
        """
        self.data = np.asarray(data)
        self.n = n
        self.idx = np.arange(data.shape[0])
        self.boxes = []
        self.rect = Rectangle(data.min(0), data.max(0))
        self.tree = innernode(self.n, self.idx, self.rect, self)


class innernode:
    """
    Represents a node in the NDTree. Each node either stores a bounding box for
    the data it contains (leaf nodes) or splits the data into two child nodes.

    Attributes
    ----------
    n : int
        The maximum number of points allowed in a leaf node for this subtree.
    idx : np.ndarray
        The indices of the data points in this node.
    tree : NDTree
        The reference to the main NDTree that holds the data and bounding boxes.
    rect : Rectangle
        The bounding box of the data points in this node.
    split_dim : int
        The dimension along which the node splits the data.
    split_point : float
        The value along the split dimension used to divide the data.
    less : innernode
        The child node containing data points less than or equal to the split
        point.
    greater : innernode
        The child node containing data points greater than the split point.
    """

    def __init__(self, n, idx, rect, tree):
        """
        Initializes the innernode and splits the data if necessary.
        """
        self.n = n
        self.idx = idx
        self.tree = tree
        self.rect = rect
        if not n == 1:
            self.split()
        else:
            box = shapely.box(*self.rect.mins, *self.rect.maxes)
            self.tree.boxes.append(box)

    def split(self):
        """
        Recursively splits the node's data into two child nodes along the
        dimension with the largest spread.
        """
        less = math.floor(self.n // 2)
        greater = self.n - less
        data = self.tree.data[self.idx]
        self.split_dim = np.argmax(self.rect.maxes - self.rect.mins)
        data = data[:, self.split_dim]
        self.split_point = np.quantile(data, less / (less + greater))
        mask = data <= self.split_point
        less_rect, greater_rect = self.rect.split(self.split_dim, self.split_point)
        self.less = innernode(less, self.idx[mask], less_rect, self.tree)
        self.greater = innernode(greater, self.idx[~mask], greater_rect, self.tree)



import torch.nn
import torch.nn.functional as F
from torch import Tensor, LongTensor
from sklearn.preprocessing import LabelEncoder
from typing import Optional, Union
from numpy.typing import ArrayLike
import pandas as pd


# TODO: Add documentation
class TranscriptEmbedding(torch.nn.Module):
    """
    Utility class to handle transcript embeddings in PyTorch so that they are
    optionally learnable in the future.

    Default behavior is to use the index of gene names.
    """

    # TODO: Add documentation
    @staticmethod
    def _check_inputs(
        classes: ArrayLike,
        weights: Union[pd.DataFrame, None],
    ):
        # Classes is a 1D array
        if len(classes.shape) > 1:
            msg = "'classes' should be a 1D array, got an array of shape " f"{classes.shape} instead."
            raise ValueError(msg)
        # Items appear exactly once
        if len(classes) != len(set(classes)):
            msg = "All embedding classes must be unique. One or more items in " "'classes' appears twice."
            raise ValueError(msg)
        # All classes have an entry in weights
        elif weights is not None:
            missing = set(classes).difference(weights.index)
            if len(missing) > 0:
                msg = f"Index of 'weights' DataFrame is missing {len(missing)} " "entries compared to classes."
                raise ValueError(msg)

    # TODO: Add documentation
    def __init__(
        self,
        classes: ArrayLike,
        weights: Optional[pd.DataFrame] = None,
    ):
        # check input arguments
        TranscriptEmbedding._check_inputs(classes, weights)
        # Setup as PyTorch module
        super(TranscriptEmbedding, self).__init__()
        self._encoder = LabelEncoder().fit(classes)
        if weights is None:
            self._weights = None
        else:
            self._weights = Tensor(weights.loc[classes].values)

    # TODO: Add documentation
    def embed(self, classes: ArrayLike):
        indices = LongTensor(self._encoder.transform(classes))
        # Default, one-hot encoding
        if self._weights is None:
            return indices  # F.one_hot(indices, len(self._encoder.classes_))
        else:
            return F.embedding(indices, self._weights)



import click
from stereosegger.cli.utils import add_options, CustomFormatter
from pathlib import Path
import logging
from argparse import Namespace

# Path to default YAML configuration file
train_yml = Path(__file__).parent / "configs" / "train" / "default.yaml"

help_msg = "Train the Segger segmentation model."


@click.command(name="train_model", help=help_msg)
@add_options(config_path=train_yml)
@click.option("--dataset_dir", type=Path, required=True, help="Directory containing the processed Segger dataset.")
@click.option(
    "--models_dir", type=Path, required=True, help="Directory to save the trained model and the training logs."
)
@click.option("--sample_tag", type=str, required=True, help="Sample tag for the dataset.")
@click.option("--init_emb", type=int, default=8, help="Size of the embedding layer.")
@click.option("--hidden_channels", type=int, default=32, help="Size of hidden channels in the model.")
@click.option("--num_tx_tokens", type=int, default=500, help="Number of transcript tokens.")
@click.option("--out_channels", type=int, default=8, help="Number of output channels.")
@click.option("--heads", type=int, default=2, help="Number of attention heads.")
@click.option("--num_mid_layers", type=int, default=2, help="Number of mid layers in the model.")
@click.option("--batch_size", type=int, default=4, help="Batch size for training.")
@click.option("--num_workers", type=int, default=4, help="Number of workers for data loading.")
@click.option(
    "--accelerator", type=str, default="cuda", help='Device type to use for training (e.g., "cuda", "cpu").'
)  # Ask for accelerator
@click.option("--max_epochs", type=int, default=200, help="Number of epochs for training.")
@click.option("--save_best_model", type=bool, default=True, help="Whether to save the best model.")  # unused for now
@click.option("--learning_rate", type=float, default=1e-3, help="Learning rate for training.")
@click.option(
    "--pretrained_model_dir",
    type=Path,
    default=None,
    help="Directory containing the pretrained modelDirectory containing the pretrained model to use (if any).",
)
@click.option("--pretrained_model_version", type=int, default=None, help="Version of pretrained model.")
@click.option("--devices", type=int, default=4, help="Number of devices (GPUs) to use.")
@click.option("--strategy", type=str, default="auto", help="Training strategy for the trainer.")
@click.option("--precision", type=str, default="16-mixed", help="Precision for training.")
def train_model(args: Namespace):

    # Setup logging
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(CustomFormatter())
    logging.basicConfig(level=logging.INFO, handlers=[ch])

    # Import packages
    logging.info("Importing packages...")
    import torch
    from stereosegger.training.train import LitSegger
    from stereosegger.training.segger_data_module import SeggerDataModule
    from stereosegger.prediction.predict_parquet import load_model
    from lightning.pytorch.loggers import CSVLogger
    from pytorch_lightning import Trainer

    logging.info("Done.")

    # Load datasets
    logging.info("Loading Xenium datasets...")
    dm = SeggerDataModule(
        data_dir=args.dataset_dir,
        batch_size=args.batch_size,  # Hard-coded batch size
        num_workers=args.num_workers,  # Hard-coded number of workers
    )

    dm.setup()
    logging.info("Done.")

    # Initialize model
    if args.pretrained_model_dir is not None:
        logging.info("Loading pretrained model...")
        ls = load_model(args.pretrained_model_dir / "lightning_logs" / f"version_{args.model_version}" / "checkpoints")
    else:
        logging.info("Creating new model...")
        token_flag = getattr(dm.train[0]["tx"], "token_based", None)
        if token_flag is not None:
            if torch.is_tensor(token_flag):
                is_token_based = bool(token_flag.item())
            else:
                is_token_based = bool(token_flag)
        else:
            is_token_based = dm.train[0].x_dict["tx"].ndim == 1
        if is_token_based:
            # if the model is token-based, the input is a 1D tensor of token indices
            tx_x = dm.train[0].x_dict["tx"]
            assert tx_x.ndim in (1, 2)
            if tx_x.ndim == 1:
                assert tx_x.dtype == torch.long
            else:
                assert tx_x.shape[1] == 2
            num_tx_features = args.num_tx_tokens
            print("Using token-based embeddings as node features, number of tokens: ", num_tx_features)
        else:
            # if the model is not token-based, the input is a 2D tensor of scRNAseq embeddings
            assert dm.train[0].x_dict["tx"].ndim == 2
            assert dm.train[0].x_dict["tx"].dtype == torch.float32
            num_tx_features = dm.train[0].x_dict["tx"].shape[1]
            print("Using scRNAseq embeddings as node features, number of features: ", num_tx_features)
        num_bd_features = dm.train[0].x_dict["bd"].shape[1]
        print("Number of boundary node features: ", num_bd_features)
        ls = LitSegger(
            is_token_based=is_token_based,
            num_node_features={"tx": num_tx_features, "bd": num_bd_features},
            init_emb=args.init_emb,
            hidden_channels=args.hidden_channels,
            out_channels=args.out_channels,
            heads=args.heads,
            num_mid_layers=args.num_mid_layers,
            aggr="sum",  # Hard-coded value
            learning_rate=args.learning_rate,
        )
    logging.info("Done.")

    # Initialize the Lightning trainer
    trainer = Trainer(
        accelerator=args.accelerator,  # Directly use the specified accelerator
        strategy=args.strategy,  # Hard-coded value
        precision=args.precision,  # Hard-coded value
        devices=args.devices,  # Hard-coded value
        max_epochs=args.max_epochs,  # Hard-coded value
        default_root_dir=args.models_dir,
        logger=CSVLogger(args.models_dir),
    )

    logging.info("Done.")

    # Train model
    logging.info("Training model...")
    trainer.fit(model=ls, datamodule=dm)
    logging.info("Done.")


@click.command(name="slurm", help="Train on Slurm cluster")
@add_options(config_path=train_yml)
def train_slurm(args):
    train_model(args)


@click.group(help="Train the Segger model")
def train():
    pass


train.add_command(train_slurm)

if __name__ == "__main__":
    train_model()



"""
Prediction module for Segger.

Contains prediction scripts and utilities for the Segger model.
"""

__all__ = ["load_model", "segment", "predict_batch"]

try:
    from .predict_parquet import load_model, segment, predict_batch
except ImportError:
    # Handle missing dependencies (like cupy)
    def load_model(*args, **kwargs):
        raise ImportError("load_model requires cupy, which is not installed.")
    def segment(*args, **kwargs):
        raise ImportError("segment requires cupy, which is not installed.")
    def predict_batch(*args, **kwargs):
        raise ImportError("predict_batch requires cupy, which is not installed.")



import os
import torch
import cupy as cp
import pandas as pd
import numpy as np
import torch.nn.functional as F
import torch._dynamo
import gc

# import rmm
import re
import glob
from pathlib import Path
from torch_geometric.loader import DataLoader
from torch_geometric.data import Batch
from stereosegger.data.utils import (
    get_edge_index_cuda,
    get_edge_index,
    format_time,
    create_anndata,
    coo_to_dense_adj,
)
from stereosegger.training.train import LitSegger
from stereosegger.training.segger_data_module import SeggerDataModule
from scipy.sparse.csgraph import connected_components as cc
from typing import Union, Dict
import dask.dataframe as dd
from dask import delayed
from dask.diagnostics import ProgressBar
import time
from cupyx.scipy.sparse import coo_matrix
from scipy.sparse import coo_matrix as scipy_coo_matrix
from torch.utils.dlpack import to_dlpack


def _get_id():
    """Generate a random Xenium-style ID."""
    return "".join(np.random.choice(list("abcdefghijklmnopqrstuvwxyz"), 8)) + "-nx"


# CONFIG
torch._dynamo.config.suppress_errors = True
os.environ["PYTORCH_USE_CUDA_DSA"] = "1"
os.environ["CUDA_LAUNCH_BLOCKING"] = "1"


def load_model(checkpoint_path: str) -> LitSegger:
    """
    Load a LitSegger model from a checkpoint.

    Parameters
    ----------
    checkpoint_path : str
        Specific checkpoint file to load, or directory where the model checkpoints are stored.
        If directory, the latest checkpoint is loaded.

    Returns
    -------
    LitSegger
        The loaded LitSegger model.

    Raises
    ------
    FileNotFoundError
        If the specified checkpoint file does not exist.
    """
    checkpoint_path = Path(checkpoint_path)
    msg = f"No checkpoint found at {checkpoint_path}. Please make sure you've provided the correct path."

    # Get last checkpoint if directory is provided
    if os.path.isdir(checkpoint_path):
        checkpoints = glob.glob(str(checkpoint_path / "*.ckpt"))
        if len(checkpoints) == 0:
            raise FileNotFoundError(msg)

        # Sort checkpoints by epoch and step
        def sort_order(c):
            match = re.match(r".*epoch=(\d+)-step=(\d+).ckpt", c)
            return int(match[1]), int(match[2])

        checkpoint_path = Path(sorted(checkpoints, key=sort_order)[-1])
    elif not checkpoint_path.exists():
        raise FileExistsError(msg)

    # Load model from checkpoint
    lit_segger = LitSegger.load_from_checkpoint(
        checkpoint_path=checkpoint_path,
    )

    return lit_segger


def get_similarity_scores(
    model: torch.nn.Module,
    batch: Batch,
    from_type: str,
    to_type: str,
    receptive_field: dict,
    knn_method: str,
) -> coo_matrix:
    """
    Compute similarity scores between embeddings for 'from_type' and 'to_type' nodes
    using sparse matrix multiplication with CuPy and the 'sees' edge relation.

    Args:
        model (torch.nn.Module): The segmentation model used to generate embeddings.
        batch (Batch): A batch of data containing input features and edge indices.
        from_type (str): The type of node from which the similarity is computed.
        to_type (str): The type of node to which the similarity is computed.
        knn_method (str, optional): The method to use for nearest neighbors. Defaults to 'cuda'.

    Returns:
        coo_matrix: A sparse matrix containing the similarity scores between
                    'from_type' and 'to_type' nodes.
    """

    # Keep everything on GPU until final results
    batch = batch.to("cuda")

    # Step 1: Get embeddings from the model (on GPU)
    shape = batch[from_type].x.shape[0], batch[to_type].x.shape[0]

    # Compute edge indices using knn method (still on GPU)
    edge_index = get_edge_index(
        batch[to_type].pos[:, :2],  # 'tx' positions
        batch[from_type].pos[:, :2],  # 'bd' positions
        k=receptive_field[f"k_{to_type}"],
        dist=receptive_field[f"dist_{to_type}"],
        method=knn_method,
    )

    # Convert to dense adjacency matrix (on GPU)
    edge_index = coo_to_dense_adj(
        edge_index.T,
        num_nodes=shape[0],
        num_nbrs=receptive_field[f"k_{to_type}"],
    )

    with torch.no_grad():
        embeddings = model(batch.x_dict, batch.edge_index_dict)

    del batch

    # print(edge_index)
    # print(embeddings)

    def sparse_multiply(embeddings, edge_index, shape) -> coo_matrix:
        m = torch.nn.ZeroPad2d((0, 0, 0, 1))  # pad bottom with zeros

        similarity = torch.bmm(
            m(embeddings[to_type])[edge_index],  # 'to' x 'from' neighbors x embed
            embeddings[from_type].unsqueeze(-1),  # 'to' x embed x 1
        )  # -> 'to' x 'from' neighbors x 1
        del embeddings
        # Sigmoid to get most similar 'to_type' neighbor
        similarity[similarity == 0] = -torch.inf  # ensure zero stays zero
        similarity = F.sigmoid(similarity)
        # Neighbor-filtered similarity scores
        # shape = batch[from_type].x.shape[0], batch[to_type].x.shape[0]
        indices = torch.argwhere(edge_index != -1).T
        indices[1] = edge_index[edge_index != -1]
        rows = cp.fromDlpack(to_dlpack(indices[0, :].to("cuda")))
        columns = cp.fromDlpack(to_dlpack(indices[1, :].to("cuda")))
        # print(rows)
        del indices
        values = similarity[edge_index != -1].flatten()
        sparse_result = coo_matrix((cp.fromDlpack(to_dlpack(values)), (rows, columns)), shape=shape)
        return sparse_result
        # Free GPU memory after computation

    # Call the sparse multiply function
    sparse_similarity = sparse_multiply(embeddings, edge_index, shape)

    return sparse_similarity


def predict_batch(
    lit_segger: torch.nn.Module,
    batch: Batch,
    score_cut: float,
    receptive_field: Dict[str, float],
    use_cc: bool = True,
    knn_method: str = "cuda",
) -> pd.DataFrame:
    """
    Predict cell assignments for a batch of transcript data using a segmentation model.
    Adds a 'bound' column to indicate if the transcript is assigned to a cell (bound=1)
    or unassigned (bound=0). Unassigned transcripts are handled with connected components
    if use_cc is True.
    Adds a 'bound' column to indicate if the transcript is assigned to a cell (bound=1)
    or unassigned (bound=0).

    Args:
        lit_segger (torch.nn.Module): The lightning module wrapping the segmentation model.
        batch (Batch): A batch of transcript and cell data.
        score_cut (float): The threshold for assigning transcripts to cells based on similarity scores.
        receptive_field (Dict[str, float]): Dictionary defining the receptive field for transcript-cell
                                            and transcript-transcript relations.
        use_cc (bool, optional): If True, perform connected components analysis for unassigned transcripts.
                                 Defaults to True.
        knn_method (str, optional): The method to use for nearest neighbors. Defaults to 'cuda'.

    Returns:
        pd.DataFrame: A DataFrame containing the transcript IDs, similarity scores,
                      assigned cell IDs, and 'bound' column.
    """

    def _get_id():
        """Generate a random Xenium-style ID."""
        return "".join(np.random.choice(list("abcdefghijklmnopqrstuvwxyz"), 8)) + "-nx"

    # Use CuPy with GPU context
    with cp.cuda.Device(0):
        # Move batch to GPU
        batch = batch.to("cuda")

        # Extract transcript IDs and initialize assignments DataFrame
        transcript_id = batch["tx"].id.cpu()
        assignments = pd.DataFrame({"transcript_id": transcript_id})
        transcript_id = cp.asnumpy(batch["tx"].id)
        assignments = pd.DataFrame({"transcript_id": transcript_id})

        if len(batch["bd"].pos) >= 10:
            # Compute similarity scores between 'tx' and 'bd'
            scores = get_similarity_scores(lit_segger.model, batch, "tx", "bd", receptive_field, knn_method=knn_method)
            scores = get_similarity_scores(lit_segger.model, batch, "tx", "bd", receptive_field, knn_method)
            torch.cuda.empty_cache()
            # Convert sparse matrix to dense format
            dense_scores = scores.toarray()  # Convert to dense NumPy array
            del scores  # Remove from memory
            cp.get_default_memory_pool().free_all_blocks()  # Free CuPy memory

            # Get direct assignments from similarity matrix
            belongs = cp.max(dense_scores, axis=1)  # Max score per transcript
            assignments["score"] = cp.asnumpy(belongs)  # Move back to CPU

            mask = assignments["score"] > score_cut
            all_ids = np.concatenate(batch["bd"].id)  # Keep IDs as NumPy array
            assignments["segger_cell_id"] = None  # Initialize as None
            max_indices = cp.argmax(dense_scores, axis=1).get()
            assignments.loc[mask, "segger_cell_id"] = all_ids[max_indices[mask]]

            assignments.loc[mask, "segger_cell_id"] = all_ids[max_indices[mask]]  # Assign IDs

            del dense_scores  # Remove from memory
            cp.get_default_memory_pool().free_all_blocks()  # Free CuPy memory
            torch.cuda.empty_cache()

            assignments["bound"] = 0
            assignments.loc[mask, "bound"] = 1

            # Handle unassigned transcripts with connected components
            # Move back to CPU
            assignments["bound"] = 0
            assignments.loc[mask, "bound"] = 1

            if use_cc:
                # Compute similarity scores between 'tx' and 'tx'
                scores_tx = get_similarity_scores(
                    lit_segger.model, batch, "tx", "tx", receptive_field, knn_method=knn_method
                )

                # Convert to dense NumPy array
                data_cpu = scores_tx.data.get()  # Transfer data to CPU (NumPy)
                row_cpu = scores_tx.row.get()  # Transfer row indices to CPU (NumPy)
                col_cpu = scores_tx.col.get()  # Transfer column indices to CPU (NumPy)
                # Remove from memory
                scores_tx = get_similarity_scores(
                    lit_segger.model, batch, "tx", "tx", receptive_field, knn_method=knn_method
                )
                # Convert to dense NumPy array
                data_cpu = scores_tx.data.get()  # Transfer data to CPU (NumPy)
                row_cpu = scores_tx.row.get()  # Transfer row indices to CPU (NumPy)
                col_cpu = scores_tx.col.get()  # Transfer column indices to CPU (NumPy)

                # dense_scores_tx = scores_tx.toarray().astype(cp.float16)
                # Rebuild the matrix on CPU using SciPy
                dense_scores_tx = scipy_coo_matrix((data_cpu, (row_cpu, col_cpu)), shape=scores_tx.shape).toarray()
                del scores_tx
                np.fill_diagonal(dense_scores_tx, 0)  # Ignore self-similarity

                # Get the indices of unassigned transcripts
                no_id = assignments["segger_cell_id"].isna().values  # Convert to numpy for indexing

                del scores_tx  # Remove from memory
                cp.get_default_memory_pool().free_all_blocks()  # Free CuPy memory

                # Assign unassigned transcripts using connected components
                no_id = assignments["segger_cell_id"].isna()
                if np.any(no_id):  # Only compute if there are unassigned transcripts
                    # Transfer the relevant parts of the sparse matrix (unassigned transcripts)
                    no_id_scores = dense_scores_tx[no_id][:, no_id]

                    # Apply score cut-off
                    # no_id_scores = no_id_scores.toarray()  # Move to dense temporarily
                    no_id_scores[no_id_scores < score_cut] = 0  # Apply threshold
                    no_id_scores = scipy_coo_matrix(no_id_scores)  # Convert back to sparse

                    # Find the non-zero entries in the no_id_scores to construct edge_index
                    non_zero_rows, non_zero_cols = no_id_scores.nonzero()

                    # Map these indices back to the actual transcript IDs (no_id_mask gives us their original position)
                    unassigned_ids = batch["tx"].id[no_id]  # Unassigned transcript IDs

                    # Construct edge index (source, target) based on non-zero connections in the no_id_scores matrix
                    source_nodes = unassigned_ids[non_zero_rows].cpu()
                    target_nodes = unassigned_ids[non_zero_cols].cpu()

                    # Convert to Dask array for later concatenation
                    edge_index = dd.from_array(np.stack([source_nodes, target_nodes], axis=0))

                    # Append this batch's edges to a global list to be used later in segment
                    global_edge_index_list.append(edge_index)
                    del dense_scores_tx  # Remove from memory
                    no_id_scores[no_id_scores < score_cut] = 0
                    n, comps = cc(no_id_scores, connection="weak", directed=False)
                    new_ids = np.array([_get_id() for _ in range(n)])
                    assignments["segger_cell_id"][no_id] = new_ids[comps]

                    # Free memory after computation
                    del no_id_scores
                    cp.get_default_memory_pool().free_all_blocks()
                    torch.cuda.empty_cache()

        return assignments


def predict(
    lit_segger: LitSegger,
    data_loader: DataLoader,
    score_cut: float,
    receptive_field: dict,
    use_cc: bool = True,
    knn_method: str = "cuda",
) -> pd.DataFrame:  # Change return type to Dask DataFrame if applicable
    """
    Optimized prediction for multiple batches of transcript data, using parallel execution
    and displaying a progress bar.

    Args:
        lit_segger (LitSegger): The trained segmentation model.
        data_loader (DataLoader): PyTorch DataLoader to iterate over transcript data batches.
        score_cut (float): Threshold for assigning transcripts to cells based on similarity scores.
        receptive_field (dict): Defines the receptive field for transcript-cell and transcript-transcript relations.
        use_cc (bool, optional): If True, perform connected components analysis for unassigned transcripts. Defaults to True.
        knn_method (str, optional): Method to use for nearest neighbors ('cuda', 'kd_tree', etc.). Defaults to 'cuda'.
        n_jobs (int, optional): Number of parallel jobs to run. Defaults to 4.

    Returns:
        pd.DataFrame: Final DataFrame containing the transcript IDs, similarity scores, and assigned cell IDs.
    """

    all_assignments = []
    # Use Dask delayed to parallelize the predict_batch execution and create assignments
    delayed_batches = [
        delayed(predict_batch)(lit_segger, batch, score_cut, receptive_field, use_cc, knn_method)
        for batch in data_loader
    ]

    for batch in data_loader:
        assignments = predict_batch(lit_segger, batch, score_cut, receptive_field, use_cc, knn_method)
        all_assignments.append(dd.from_pandas(assignments, npartitions=1))

        cp.get_default_memory_pool().free_all_blocks()
        torch.cuda.empty_cache()

    # Concatenate all assignments into a single Dask DataFrame
    final_assignments = dd.concat(all_assignments, ignore_index=True)

    # Max score selection logic
    max_bound_idx = final_assignments[final_assignments["bound"] == 1].groupby("transcript_id")["score"].idxmax()
    max_unbound_idx = final_assignments[final_assignments["bound"] == 0].groupby("transcript_id")["score"].idxmax()

    # Combine indices, prioritizing bound=1 scores
    final_idx = max_bound_idx.combine_first(max_unbound_idx).compute()  # Ensure it's computed

    # Now use the computed final_idx for indexing
    result = final_assignments.loc[final_idx].compute().reset_index(drop=True)

    # result = results.reset_index()

    # Handle cases where there's only one entry per 'segger_cell_id'
    # single_entry_mask = result.groupby('segger_cell_id').size() == 1
    # Handle cases where there's only one entry per 'segger_cell_id'
    # single_entry_counts = result['segger_cell_id'].value_counts()  # Count occurrences of each ID
    # single_entry_mask = single_entry_counts[single_entry_counts == 1].index  # Get IDs with a count of 1

    # # Update 'segger_cell_id' for single entries
    # for segger_id in single_entry_mask:
    #     result.loc[result['segger_cell_id'] == segger_id, 'segger_cell_id'] = 'floating'

    return result


def segment(
    model: LitSegger,
    dm: SeggerDataModule,
    save_dir: Union[str, Path],
    seg_tag: str,
    transcript_file: Union[str, Path],
    score_cut: float = 0.5,
    use_cc: bool = True,
    file_format: str = "anndata",
    receptive_field: dict = {"k_bd": 4, "dist_bd": 10, "k_tx": 5, "dist_tx": 3},
    knn_method: str = "kd_tree",
    verbose: bool = False,
    **anndata_kwargs,
) -> None:
    """
    Perform segmentation using the model, merge segmentation results with transcripts_df, and save in the specified format.

    Parameters:
    ----------
    model : LitSegger
        The trained segmentation model.
    dm : SeggerDataModule
        The SeggerDataModule instance for data loading.
    save_dir : Union[str, Path]
        Directory to save the final segmentation results.
    seg_tag : str
        Tag to include in the saved filename.
    transcript_file : Union[str, Path]
        Path to the transcripts parquet file.
    file_format : str, optional
        File format to save the results ('csv', 'parquet', or 'anndata'). Defaults to 'anndata'.
    score_cut : float, optional
        The threshold for assigning transcripts to cells based on similarity scores.
    use_cc : bool, optional
        If to further re-group transcripts that have not been assigned to any nucleus.
    knn_method : str, optional
        The method to use for nearest neighbors ('cuda' by default).
    **anndata_kwargs : dict, optional
        Additional keyword arguments passed to the create_anndata function.
    Perform segmentation using the model, merge segmentation results with transcripts data,
    and save the results in the specified format. The function also handles unassigned
    transcripts through connected components analysis if `use_cc` is True.

    Args:
        model (LitSegger): The trained segmentation model.
        dm (SeggerDataModule): The SeggerDataModule instance for data loading.
        save_dir (Union[str, Path]): Directory to save the final segmentation results.
        seg_tag (str): Tag to include in the saved filename.
        transcript_file (Union[str, Path]): Path to the transcripts Parquet file.
        score_cut (float, optional): The threshold for assigning transcripts to cells based on
                                     similarity scores. Defaults to 0.5.
        use_cc (bool, optional): If True, perform connected components analysis for unassigned
                                 transcripts. Defaults to True.
        file_format (str, optional): The file format to save the results ('csv', 'parquet', or 'anndata').
                                     Defaults to 'anndata'.
        receptive_field (dict, optional): Defines the receptive field for transcript-cell and
                                          transcript-transcript relations. Defaults to
                                          {'k_bd': 4, 'dist_bd': 10, 'k_tx': 5, 'dist_tx': 3}.
        knn_method (str, optional): The method to use for nearest neighbors ('cuda' or 'kd_tree').
                                    Defaults to 'kd_tree'.
        verbose (bool, optional): Whether to print verbose status updates. Defaults to False.
        **anndata_kwargs: Additional keyword arguments passed to the `create_anndata` function.

    Returns:
        None
    """

    # Initialize global_edge_index_list to store edge indices across batches
    global global_edge_index_list
    global_edge_index_list = []

    # Start the timer
    start_time = time.time()

    # Ensure the save directory exists
    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    if verbose:
        print(f"Starting segmentation for {seg_tag}...")

    # Step 1: Prediction
    step_start_time = time.time()

    train_dataloader = dm.train_dataloader()
    test_dataloader = dm.test_dataloader()
    val_dataloader = dm.val_dataloader()

    segmentation_train = predict(model, train_dataloader, score_cut, receptive_field, use_cc, knn_method)
    torch.cuda.empty_cache()
    cp.get_default_memory_pool().free_all_blocks()
    gc.collect()

    segmentation_val = predict(model, val_dataloader, score_cut, receptive_field, use_cc, knn_method)
    torch.cuda.empty_cache()
    cp.get_default_memory_pool().free_all_blocks()
    gc.collect()

    segmentation_test = predict(model, test_dataloader, score_cut, receptive_field, use_cc, knn_method)
    torch.cuda.empty_cache()
    cp.get_default_memory_pool().free_all_blocks()
    gc.collect()

    if verbose:
        elapsed_time = format_time(time.time() - step_start_time)
        print(f"Predictions completed in {elapsed_time}.")

    # Step 2: Combine and group by transcript_id
    step_start_time = time.time()

    # Combine the segmentation data
    seg_combined = pd.concat([segmentation_train, segmentation_val, segmentation_test], ignore_index=True)

    # Drop any unassigned rows
    seg_final = seg_combined.dropna(subset=["segger_cell_id"]).reset_index(drop=True)

    if verbose:
        elapsed_time = format_time(time.time() - step_start_time)
        print(f"Segmentation results processed in {elapsed_time}.")

    # Step 3: Load transcripts and merge
    step_start_time = time.time()

    transcripts_df = dd.read_parquet(transcript_file)

    if verbose:
        print("Merging segmentation results with transcripts...")

    # Convert the segmentation results to a Dask DataFrame, keeping npartitions consistent
    seg_final_dd = dd.from_pandas(seg_final, npartitions=transcripts_df.npartitions)

    # Merge the segmentation results with the transcript data (still as Dask DataFrame)
    transcripts_df_filtered = transcripts_df.merge(seg_final_dd, on="transcript_id", how="inner")

    if verbose:
        elapsed_time = format_time(time.time() - step_start_time)
        print(f"Transcripts merged in {elapsed_time}.")

    # Step 4: Handle unassigned transcripts using connected components (if use_cc is True)
    if use_cc:
        if verbose:
            print(f"Computing connected components for unassigned transcripts...")

        # Concatenate all edge indices stored across batches
        concatenated_edge_index = dd.concat(global_edge_index_list)

        # Compute connected components using the concatenated edge_index
        edge_index_computed = concatenated_edge_index.compute()  # Get the full edge_index (source, target)

        # Map transcript_ids to their index positions in the DataFrame
        transcript_idx_map = pd.Series(
            transcripts_df_filtered.index, index=transcripts_df_filtered["transcript_id"]
        ).to_dict()

        # Convert the transcript_ids in edge_index_computed to positional indices
        source_indices = [transcript_idx_map[tid] for tid in edge_index_computed[0]]
        target_indices = [transcript_idx_map[tid] for tid in edge_index_computed[1]]

        # Use SciPy's connected components algorithm
        n, comps = cc(
            scipy_coo_matrix(
                (np.ones(len(source_indices)), (source_indices, target_indices)),
                shape=(transcripts_df_filtered.shape[0], transcripts_df_filtered.shape[0]),
            ),
            connection="weak",
            directed=False,
        )

        # Generate new cell IDs based on connected components
        new_ids = np.array([_get_id() for _ in range(n)])

        # Assign new cell IDs to the unassigned transcripts in the final assignments
        unassigned_mask = transcripts_df_filtered["segger_cell_id"].isna()
        transcripts_df_filtered.loc[unassigned_mask, "segger_cell_id"] = new_ids[comps]

        if verbose:
            elapsed_time = format_time(time.time() - step_start_time)
            print(f"Connected components computed in {elapsed_time}.")

    # Step 5: Save the merged result
    step_start_time = time.time()

    if verbose:
        print(f"Saving results in {file_format} format...")

    if file_format == "csv":
        save_path = save_dir / f"{seg_tag}_segmentation.csv"
        transcripts_df_filtered.compute().to_csv(save_path, index=False)  # Use pandas after computing
    elif file_format == "parquet":
        save_path = save_dir / f"{seg_tag}_segmentation.parquet"
        transcripts_df_filtered.to_parquet(save_path, index=False)  # Dask handles Parquet fine
    elif file_format == "anndata":
        save_path = save_dir / f"{seg_tag}_segmentation.h5ad"
        segger_adata = create_anndata(transcripts_df_filtered.compute(), **anndata_kwargs)  # Compute for AnnData
        segger_adata.write(save_path)
    else:
        raise ValueError(f"Unsupported file format: {file_format}")

    if verbose:
        elapsed_time = format_time(time.time() - step_start_time)
        print(f"Results saved in {elapsed_time} at {save_path}.")

    # Total time
    if verbose:
        total_time = format_time(time.time() - start_time)
        print(f"Total segmentation process completed in {total_time}.")

    # Step 6: Garbage collection and memory cleanup
    torch.cuda.empty_cache()
    gc.collect()



import torch
from torch_geometric.nn import HeteroConv, GATv2Conv, HeteroDictLinear
from torch import nn
import torch.nn.functional as F
from torch import Tensor
from typing import Union


class SkipGAT(nn.Module):
    def __init__(self, in_channels, out_channels, heads, apply_activation=True):
        super().__init__()
        self.apply_activation = apply_activation
        self.conv = HeteroConv(
            {
                ("tx", "neighbors", "tx"): GATv2Conv(in_channels, out_channels, heads=heads),
                ("tx", "belongs", "bd"): GATv2Conv(in_channels, out_channels, heads=heads, add_self_loops=False),
            },
            aggr="sum",
        )
        self.lin = HeteroDictLinear(in_channels, out_channels * heads, types=("tx", "bd"))

    def forward(self, x_dict, edge_index_dict):
        x_conv = self.conv(x_dict, edge_index_dict)
        x_lin = self.lin(x_dict)
        x_dict = {key: x_conv[key] + x_lin[key] for key in x_dict}
        if self.apply_activation:
            x_dict = {key: F.leaky_relu(x) for key, x in x_dict.items()}
        return x_dict


class Segger(nn.Module):
    def __init__(
        self,
        is_token_based: int,
        num_node_features: dict[str, int],
        init_emb: int = 16,
        hidden_channels: int = 32,
        num_mid_layers: int = 3,
        out_channels: int = 32,
        heads: int = 3,
    ):
        """
        Initializes the Segger model.

        Args:
            is_token_based (int)   : Whether the model is using token-based embeddings or scRNAseq embeddings.
            num_node_features (dict[str, int]): Number of node features for each node type.
            init_emb (int)         : Initial embedding size for both 'tx' and boundary (non-token) nodes.
            hidden_channels (int)  : Number of hidden channels.
            num_mid_layers (int)   : Number of hidden layers (excluding first and last layers).
            out_channels (int)     : Number of output channels.
            heads (int)            : Number of attention heads.
        """
        super().__init__()
        self.is_token_based = bool(is_token_based)

        # Initialize node embeddings
        if is_token_based:
            # Using token-based embeddings for transcript ('tx') nodes
            self.node_init = nn.ModuleDict(
                {
                    "tx": nn.Embedding(num_node_features["tx"], init_emb),
                    "bd": nn.Linear(num_node_features["bd"], init_emb),
                }
            )
        else:
            # Using scRNAseq embeddings (i.e. prior biological knowledge) for transcript ('tx') nodes
            self.node_init = nn.ModuleDict(
                {
                    "tx": nn.Linear(num_node_features["tx"], init_emb),
                    "bd": nn.Linear(num_node_features["bd"], init_emb),
                }
            )

        # First GATv2Conv layer
        self.conv1 = SkipGAT(init_emb, hidden_channels, heads)

        # Middle GATv2Conv layers
        self.num_mid_layers = num_mid_layers
        if num_mid_layers > 0:
            self.conv_mid_layers = nn.ModuleList()
            for _ in range(num_mid_layers):
                self.conv_mid_layers.append(SkipGAT(heads * hidden_channels, hidden_channels, heads))

        # Last GATv2Conv layer
        self.conv_last = SkipGAT(heads * hidden_channels, out_channels, heads)

        # Finalize node embeddings
        self.node_final = HeteroDictLinear(heads * out_channels, out_channels, types=("tx", "bd"))

        # # Edge probability predictor
        # self.edge_predictor = nn.Sequential(
        #     nn.LeakyReLU(),
        #     nn.Linear(2 * out_channels, out_channels),
        #     nn.LeakyReLU(),
        #     nn.Linear(out_channels, 1),
        # )

    def init_node_features(self, x_dict: dict[str, Tensor]) -> dict[str, Tensor]:
        if self.is_token_based:
            tx_input = x_dict["tx"]
            tx_count = None
            if tx_input.ndim == 2 and tx_input.shape[1] == 2:
                tx_tokens = tx_input[:, 0].long()
                tx_count = tx_input[:, 1].float().unsqueeze(1)
            else:
                tx_tokens = tx_input
            x_dict = {
                "tx": self.node_init["tx"](tx_tokens),
                "bd": self.node_init["bd"](x_dict["bd"]),
            }
            if tx_count is not None:
                x_dict["tx"] = x_dict["tx"] * (1.0 + tx_count)
        else:
            x_dict = {key: self.node_init[key](x) for key, x in x_dict.items()}
        return x_dict

    def forward(
        self,
        x_dict: dict[str, Tensor],
        edge_index_dict: dict[str, Tensor],
    ) -> dict[str, Tensor]:
        """
        Forward pass for the Segger model.

        Args:
            x_dict (dict[str, Tensor]): Node features for each node type.
            edge_index_dict (dict[str, Tensor]): Edge indices for each edge type.
        """

        x_dict = self.init_node_features(x_dict)

        x_dict = {key: F.leaky_relu(x) for key, x in x_dict.items()}

        x_dict = self.conv1(x_dict, edge_index_dict)

        if self.num_mid_layers > 0:
            for i in range(self.num_mid_layers):
                x_dict = self.conv_mid_layers[i](x_dict, edge_index_dict)

        x_dict = self.conv_last(x_dict, edge_index_dict)

        x_dict = self.node_final(x_dict)

        return x_dict

    def decode(
        self,
        z_dict: dict[str, Tensor],
        edge_index: Union[Tensor],
    ) -> Tensor:
        """
        Decode the node embeddings to predict edge values.

        Args:
            z_dict (dict[str, Tensor]): Node embeddings for each node type.
            edge_index (EdgeIndex): Edge label indices.

        Returns:
            Tensor: Predicted edge values.
        """
        z_left = z_dict["tx"][edge_index[0]]
        z_right = z_dict["bd"][edge_index[1]]
        return (z_left * z_right).sum(dim=-1)
        # return self.edge_predictor(torch.cat([z_left, z_right], dim=-1)).squeeze()



from typing import List, Optional, Callable
from torch_geometric.data import Dataset, Data
import glob
import os
from pathlib import Path
import torch


class STPyGDataset(Dataset):
    """
    A dataset class for handling training using spatial
    transcriptomics data, loading from disk.
    """

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
    ):
        super().__init__(root, transform, pre_transform, pre_filter)
        print(f"Indexing dataset at {self.processed_dir}...")
        # Faster than glob.glob for large directories
        paths = [
            f.path for f in os.scandir(self.processed_dir) 
            if f.is_file() and f.name.startswith("tiles_") and f.name.endswith(".pt")
        ]
        self._processed_file_paths = sorted(paths)
        print(f"Found {len(self._processed_file_paths)} tiles.")

    @property
    def raw_file_names(self) -> List[str]:
        """
        Return a list of raw file names in the raw directory.

        Returns:
            List[str]: List of raw file names.
        """
        return os.listdir(self.raw_dir)

    @property
    def processed_file_names(self) -> List[str]:
        """
        Return a list of processed file names in the processed directory.

        Returns:
            List[str]: List of processed file names.
        """
        return [os.path.basename(p) for p in self._processed_file_paths]

    def len(self) -> int:
        """
        Return the number of processed files.

        Returns:
            int: Number of processed files.
        """
        return len(self._processed_file_paths)

    def get(self, idx: int) -> Data:
        """
        Get a processed data object.

        Args:
            idx (int): Index of the data object to retrieve.

        Returns:
            Data: The processed data object.
        """
        data = torch.load(self._processed_file_paths[idx], weights_only=False)
        # this is an issue in PyG's RandomLinkSplit, dimensions are not consistent if there is only one edge in the graph
        if hasattr(data["tx", "belongs", "bd"], "edge_label_index"):
            if data["tx", "belongs", "bd"].edge_label_index.dim() == 1:
                data["tx", "belongs", "bd"].edge_label_index = data["tx", "belongs", "bd"].edge_label_index.unsqueeze(1)
                data["tx", "belongs", "bd"].edge_label = data["tx", "belongs", "bd"].edge_label.unsqueeze(0)
            assert data["tx", "belongs", "bd"].edge_label_index.dim() == 2
        return data



from typing import TYPE_CHECKING

if TYPE_CHECKING:  # False at runtime
    import dask, cudf, dask_cudf, pandas as pd


class BackendHandler:
    """
    A class to handle different DataFrame backends for reading and processing
    Parquet files.

    Attributes
    ----------
    _valid_backends : set
        A set of valid backend options ('pandas', 'dask', 'cudf', 'dask_cudf').
    backend : str
        The selected backend for reading Parquet files.

    Methods
    -------
    read_parquet():
        Returns the function to read Parquet files according to the selected
        backend.
    """

    _valid_backends = {
        "pandas",
        "dask",
        "cudf",
        "dask_cudf",
    }

    def __init__(self, backend):
        # Make sure requested backend is supported
        if backend in self._valid_backends:
            self.backend = backend
        else:
            valid = ", ".join(map(lambda o: f"'{o}'", self._valid_backends))
            msg = f"Unsupported backend: {backend}. Valid options are {valid}."
            raise ValueError(msg)

        # Dynamically import packages only if requested
        if self.backend == "pandas":
            import pandas as pd
        elif self.backend == "dask":
            import dask
        elif self.backend == "cudf":
            import cudf
        elif self.backend == "dask_cudf":
            import dask_cudf
        else:
            raise ValueError("Internal Error")

    @property
    def read_parquet(self):
        if self.backend == "pandas":
            return pd.read_parquet
        elif self.backend == "dask":
            return dask.dataframe.read_parquet
        elif self.backend == "cudf":
            return cudf.read_parquet
        elif self.backend == "dask_cudf":
            return dask_cudf.read_parquet
        else:
            raise ValueError("Internal Error")



import unittest
import pandas as pd
import unittest
import torch
from torch_geometric.data import Data
from segger.data.utils import *
from segger.data import XeniumSample
import unittest
import pandas as pd


class TestDataUtils(unittest.TestCase):

    def test_filter_transcripts(self):
        data = {"qv": [30, 10, 25], "feature_name": ["gene1", "NegControlProbe_gene2", "gene3"]}
        df = pd.DataFrame(data)
        filtered_df = filter_transcripts(df, min_qv=20)
        self.assertEqual(len(filtered_df), 2)
        self.assertTrue("gene1" in filtered_df["feature_name"].values)
        self.assertTrue("gene3" in filtered_df["feature_name"].values)

    def test_compute_transcript_metrics(self):
        data = {
            "qv": [40, 40, 25, 25],
            "feature_name": ["gene1", "gene2", "gene1", "gene2"],
            "cell_id": [1, 1, -1, 2],
            "overlaps_nucleus": [1, 0, 0, 1],
        }
        df = pd.DataFrame(data)
        metrics = compute_transcript_metrics(df, qv_threshold=30)
        self.assertAlmostEqual(metrics["percent_assigned"], 50.0)
        self.assertAlmostEqual(metrics["percent_cytoplasmic"], 50.0)
        self.assertAlmostEqual(metrics["percent_nucleus"], 50.0)
        self.assertAlmostEqual(metrics["percent_non_assigned_cytoplasmic"], 100.0)
        self.assertEqual(len(metrics["gene_metrics"]), 2)
        self.assertTrue("gene1" in metrics["gene_metrics"]["feature_name"].values)
        self.assertTrue("gene2" in metrics["gene_metrics"]["feature_name"].values)

    def setUp(self):
        data = {
            "x_location": [100, 200, 300],
            "y_location": [100, 200, 300],
            "z_location": [0, 0, 0],
            "qv": [40, 40, 25],
            "feature_name": ["gene1", "gene2", "gene3"],
            "transcript_id": [1, 2, 3],
            "overlaps_nucleus": [1, 0, 1],
            "cell_id": [1, -1, 2],
        }
        self.df = pd.DataFrame(data)
        self.sample = XeniumSample(self.df)

    def test_crop_transcripts(self):
        cropped_sample = self.sample.crop_transcripts(50, 50, 200, 200)
        self.assertEqual(len(cropped_sample.transcripts_df), 1)
        self.assertEqual(cropped_sample.transcripts_df.iloc[0]["feature_name"], "gene1")

    def test_filter_transcripts(self):
        filtered_df = XeniumSample.filter_transcripts(self.df, min_qv=30)
        self.assertEqual(len(filtered_df), 2)
        self.assertTrue("gene1" in filtered_df["feature_name"].values)
        self.assertTrue("gene2" in filtered_df["feature_name"].values)

    def test_unassign_all_except_nucleus(self):
        unassigned_df = XeniumSample.unassign_all_except_nucleus(self.df)
        self.assertEqual(unassigned_df.loc[unassigned_df["overlaps_nucleus"] == 0, "cell_id"].values[0], "UNASSIGNED")


if __name__ == "__main__":
    unittest.main()



import pandas as pd
import geopandas as gpd
import shapely
from pyarrow import parquet as pq
import numpy as np
import scipy as sp
from typing import Optional, List
import sys
from types import SimpleNamespace
from pathlib import Path
import yaml


def get_xy_extents(
    filepath,
    x: str,
    y: str,
) -> shapely.Polygon:
    """
    Get the bounding box of the x and y coordinates from a Parquet file.
    If the min/max statistics are not available, compute them efficiently from the data.

    Parameters
    ----------
    filepath : str
        The path to the Parquet file.
    x : str
        The name of the column representing the x-coordinate.
    y : str
        The name of the column representing the y-coordinate.

    Returns
    -------
    shapely.Polygon
        A polygon representing the bounding box of the x and y coordinates.
    """
    # Process row groups
    metadata = pq.read_metadata(filepath)
    schema_idx = dict(map(reversed, enumerate(metadata.schema.names)))
    # Find min and max values across all row groups
    x_max = -1
    x_min = sys.maxsize
    y_max = -1
    y_min = sys.maxsize
    group = metadata.row_group(0)
    try:
        for i in range(metadata.num_row_groups):
            # print('*1')
            group = metadata.row_group(i)
            x_min = min(x_min, group.column(schema_idx[x]).statistics.min)
            x_max = max(x_max, group.column(schema_idx[x]).statistics.max)
            y_min = min(y_min, group.column(schema_idx[y]).statistics.min)
            y_max = max(y_max, group.column(schema_idx[y]).statistics.max)
            bounds = shapely.box(x_min, y_min, x_max, y_max)
    # If statistics are not available, compute them manually from the data
    except:
        import gc

        print("metadata lacks the statistics of the tile's bounding box, computing might take longer!")
        parquet_file = pd.read_parquet(filepath)
        x_col = parquet_file.loc[:, x]
        y_col = parquet_file.loc[:, y]
        del parquet_file
        gc.collect()
        x_min = x_col.min()
        y_min = y_col.min()
        x_max = x_col.max()
        y_max = y_col.max()
    bounds = shapely.geometry.box(x_min, y_min, x_max, y_max)
    return bounds


def read_parquet_region(
    filepath,
    x: str,
    y: str,
    bounds: shapely.Polygon = None,
    extra_columns: list[str] = [],
    extra_filters: list[str] = [],
    row_group_chunksize: Optional[int] = None,
):
    """
    Read a region from a Parquet file based on x and y coordinates and optional
    filters.

    Parameters
    ----------
    filepath : str
        The path to the Parquet file.
    x : str
        The name of the column representing the x-coordinate.
    y : str
        The name of the column representing the y-coordinate.
    bounds : shapely.Polygon, optional
        A polygon representing the bounding box to filter the data. If None,
        no bounding box filter is applied.
    extra_columns : list of str, optional
        A list of additional columns to include in the output DataFrame.
    extra_filters : list of str, optional
        A list of additional filters to apply to the data.

    Returns
    -------
    DataFrame
        A DataFrame containing the filtered data from the Parquet file.
    """
    # Check backend and load dependencies if not already loaded

    # Find bounds of full file if not supplied
    if bounds is None:
        bounds = get_xy_extents(filepath, x, y)

    # Load pre-filtered data from Parquet file
    filters = [
        [
            (x, ">", bounds.bounds[0]),
            (y, ">", bounds.bounds[1]),
            (x, "<", bounds.bounds[2]),
            (y, "<", bounds.bounds[3]),
        ]
        + extra_filters
    ]

    columns = list({x, y} | set(extra_columns))

    region = pd.read_parquet(
        filepath,
        filters=filters,
        columns=columns,
    )
    return region


def get_polygons_from_xy(
    boundaries: pd.DataFrame,
    x: str,
    y: str,
    label: str,
) -> gpd.GeoSeries:
    """
    Convert boundary coordinates from a cuDF DataFrame to a GeoSeries of
    polygons.

    Parameters
    ----------
    boundaries : pd.DataFrame
        A DataFrame containing the boundary data with x and y coordinates
        and identifiers.
    x : str
        The name of the column representing the x-coordinate.
    y : str
        The name of the column representing the y-coordinate.
    label : str
        The name of the column representing the cell or nucleus label.


    Returns
    -------
    gpd.GeoSeries
        A GeoSeries containing the polygons created from the boundary
        coordinates.
    """
    # Polygon offsets in coords
    ids = boundaries[label].values
    splits = np.where(ids[:-1] != ids[1:])[0] + 1
    geometry_offset = np.hstack([0, splits, len(ids)])
    part_offset = np.arange(len(np.unique(ids)) + 1)

    # Convert to GeoSeries of polygons
    polygons = shapely.from_ragged_array(
        shapely.GeometryType.POLYGON,
        coords=boundaries[[x, y]],
        offsets=(geometry_offset, part_offset),
    )
    gs = gpd.GeoSeries(polygons, index=np.unique(ids))

    return gs


def filter_boundaries(
    boundaries: pd.DataFrame,
    inset: shapely.Polygon,
    outset: shapely.Polygon,
    x: str,
    y: str,
    label: str,
):
    """
    Filter boundary polygons based on their overlap with specified inset and
    outset regions.

    Parameters
    ----------
    boundaries : cudf.DataFrame
        A DataFrame containing the boundary data with x and y coordinates and
        identifiers.
    inset : shapely.Polygon
        A polygon representing the inner region to filter the boundaries.
    outset : shapely.Polygon
        A polygon representing the outer region to filter the boundaries.
    x : str
        The name of the column representing the x-coordinate.
    y : str
        The name of the column representing the y-coordinate.
    label : str
        The name of the column representing the cell or nucleus label.

    Returns
    -------
    cudf.DataFrame
        A DataFrame containing the filtered boundary polygons.

    Notes
    -----
    The function determines overlaps of boundary polygons with the specified
    inset and outset regions. It creates boolean masks for overlaps with the
    top, left, right, and bottom sides of the outset region, as well as the
    center region defined by the inset polygon. The filtering logic includes
    polygons that:
    - Are completely within the center region.
    - Overlap with the center and the left side, but not the bottom side.
    - Overlap with the center and the top side, but not the right side.
    """

    # Determine overlaps of boundary polygons
    def in_region(region):
        in_x = boundaries[x].between(region.bounds[0], region.bounds[2])
        in_y = boundaries[y].between(region.bounds[1], region.bounds[3])
        return in_x & in_y

    x1, y1, x4, y4 = outset.bounds
    x2, y2, x3, y3 = inset.bounds
    boundaries["top"] = in_region(shapely.box(x1, y1, x4, y2))
    boundaries["left"] = in_region(shapely.box(x1, y1, x2, y4))
    boundaries["right"] = in_region(shapely.box(x3, y1, x4, y4))
    boundaries["bottom"] = in_region(shapely.box(x1, y3, x4, y4))
    boundaries["center"] = in_region(inset)

    # Filter boundary polygons
    # Include overlaps with top and left, not bottom and right
    gb = boundaries.groupby(label, sort=False)
    total = gb["center"].transform("size")
    in_top = gb["top"].transform("sum")
    in_left = gb["left"].transform("sum")
    in_right = gb["right"].transform("sum")
    in_bottom = gb["bottom"].transform("sum")
    in_center = gb["center"].transform("sum")
    keep = in_center == total
    keep |= (in_center > 0) & (in_left > 0) & (in_bottom == 0)
    keep |= (in_center > 0) & (in_top > 0) & (in_right == 0)
    inset_boundaries = boundaries.loc[keep]
    return inset_boundaries


def filter_transcripts(
    transcripts_df: pd.DataFrame,
    label: Optional[str] = None,
    filter_substrings: Optional[List[str]] = None,
    min_qv: Optional[float] = None,
) -> pd.DataFrame:
    """
    Filters transcripts based on quality value and removes unwanted transcripts.

    Parameters
    ----------
    transcripts_df : pd.DataFrame
        The dataframe containing transcript data.
    label : Optional[str]
        The label of transcript features.
    filter_substrings : Optional[str]
        The list of feature substrings to remove.
    min_qv : Optional[float]
        The minimum quality value threshold for filtering transcripts.

    Returns
    -------
    pd.DataFrame
        The filtered dataframe.
    """
    mask = pd.Series(True, index=transcripts_df.index)
    if filter_substrings and label is not None and label in transcripts_df.columns:
        series = transcripts_df[label]
        if pd.api.types.is_string_dtype(series) or series.dtype == object:
            mask &= ~series.str.startswith(tuple(filter_substrings))
    if min_qv is not None and "qv" in transcripts_df.columns:
        mask &= transcripts_df["qv"].ge(min_qv)
    return transcripts_df[mask]


def load_settings(sample_type: str) -> SimpleNamespace:
    """
    Loads a matching YAML file from the _settings/ directory and converts its
    contents into a SimpleNamespace.

    Parameters
    ----------
    sample_type : str
        Name of the sample type to load (case-insensitive).

    Returns
    -------
    SimpleNamespace
        The settings loaded from the YAML file as a SimpleNamespace.

    Raises
    ------
    ValueError
        If `sample_type` does not match any filenames.
    """
    settings_dir = Path(__file__).parent.resolve() / "_settings"
    # Get a list of YAML filenames (without extensions) in the _settings dir
    filenames = [file.stem for file in settings_dir.glob("*.yaml")]
    # Convert sample_type to lowercase and check if it matches any filename
    sample_type = sample_type.lower()
    if sample_type not in filenames:
        msg = f"Sample type '{sample_type}' not found in settings. " f"Available options: {', '.join(filenames)}"
        raise FileNotFoundError(msg)
    # Load the matching YAML file
    yaml_file_path = settings_dir / f"{sample_type}.yaml"
    with yaml_file_path.open("r") as file:
        data = yaml.safe_load(file)

    # Convert the YAML data into a SimpleNamespace recursively
    return _dict_to_namespace(data)


def _dict_to_namespace(d):
    """
    Recursively converts a dictionary to a SimpleNamespace.
    """
    if isinstance(d, dict):
        d = {k: _dict_to_namespace(v) for k, v in d.items()}
        return SimpleNamespace(**d)
    return d



import os
import torch
import cupy as cp
import pandas as pd
import numpy as np
import torch.nn.functional as F
import torch._dynamo
import gc
import rmm
import re
import glob
from pathlib import Path
from torch_geometric.loader import DataLoader
from torch_geometric.data import Batch
from stereosegger.data.utils import (
    get_edge_index,
    format_time,
    create_anndata,
    coo_to_dense_adj,
)
from stereosegger.training.train import LitSegger
from stereosegger.training.segger_data_module import SeggerDataModule
from scipy.sparse.csgraph import connected_components as cc
from typing import Union, Dict
import dask.dataframe as dd
from dask import delayed
from dask.diagnostics import ProgressBar
import time
import dask
from rmm.allocators.cupy import rmm_cupy_allocator
from cupyx.scipy.sparse import coo_matrix
from torch.utils.dlpack import to_dlpack, from_dlpack

from dask.distributed import Client, LocalCluster
import cupy as cp
import numpy as np
import pandas as pd
from cupyx.scipy.sparse import coo_matrix
from cupyx.scipy.sparse import find  # To find non-zero elements in sparse matrix
from scipy.sparse.csgraph import connected_components as cc
from scipy.sparse import coo_matrix as scipy_coo_matrix
from dask.distributed import get_client
from pqdm.processes import pqdm
from tqdm import tqdm
import json
from datetime import datetime
import dask_geopandas as dgpd  # Assuming dask-geopandas is installed
import cudf
import dask_cudf
import cupy as cp
import cupyx
import warnings
import shutil
from time import time
from cupyx.scipy.sparse import coo_matrix as cp_coo_matrix
from cupyx.scipy.sparse.csgraph import connected_components as cp_cc
from dask.distributed import LocalCluster, Client, progress
import dask
import random
from dask_cuda import LocalCUDACluster

# Setup Dask cluster with 3 workers


# CONFIG
torch._dynamo.config.suppress_errors = True
os.environ["PYTORCH_USE_CUDA_DSA"] = "1"
os.environ["CUDA_LAUNCH_BLOCKING"] = "1"


# Function to zero out diagonal of sparse COO matrix
def zero_out_diagonal_gpu(sparse_matrix):
    """
    Zero out the diagonal elements of a sparse CuPy COO matrix while keeping it sparse on the GPU.

    Args:
        sparse_matrix (cupyx.scipy.sparse.coo_matrix): Input sparse matrix.

    Returns:
        cupyx.scipy.sparse.coo_matrix: Matrix with diagonal elements zeroed out.
    """
    # Filter out the diagonal (where row == col)
    non_diagonal_mask = sparse_matrix.row != sparse_matrix.col

    # Create a new sparse matrix without diagonal elements
    sparse_matrix_no_diag = cupyx.scipy.sparse.coo_matrix(
        (
            sparse_matrix.data[non_diagonal_mask],
            (sparse_matrix.row[non_diagonal_mask], sparse_matrix.col[non_diagonal_mask]),
        ),
        shape=sparse_matrix.shape,
    )

    return sparse_matrix_no_diag


# Function to subset rows and columns of a sparse matrix
def subset_sparse_matrix(sparse_matrix, row_idx, col_idx):
    """
    Subset a sparse matrix using row and column indices.

    Parameters:
    sparse_matrix (cupyx.scipy.sparse.spmatrix): The input sparse matrix in COO, CSR, or CSC format.
    row_idx (cupy.ndarray): Row indices to keep in the subset.
    col_idx (cupy.ndarray): Column indices to keep in the subset.

    Returns:
    cupyx.scipy.sparse.spmatrix: A new sparse matrix that is a subset of the input matrix.
    """

    # Convert indices to CuPy arrays if not already
    row_idx = cp.asarray(row_idx)
    col_idx = cp.asarray(col_idx)

    # Ensure sparse matrix is in COO format for easy indexing (you can use CSR/CSC if more optimal)
    sparse_matrix = sparse_matrix.tocoo()

    # Create boolean masks for the row and column indices
    row_mask = cp.isin(sparse_matrix.row, row_idx)
    col_mask = cp.isin(sparse_matrix.col, col_idx)

    # Apply masks to filter the data, row, and column arrays
    mask = row_mask & col_mask
    row_filtered = sparse_matrix.row[mask]
    col_filtered = sparse_matrix.col[mask]
    data_filtered = sparse_matrix.data[mask]

    # Map the row and col indices to the new submatrix indices
    row_mapped = cp.searchsorted(row_idx, row_filtered)
    col_mapped = cp.searchsorted(col_idx, col_filtered)

    # Return the new subset sparse matrix
    return coo_matrix((data_filtered, (row_mapped, col_mapped)), shape=(len(row_idx), len(col_idx)))


def load_model(checkpoint_path: str) -> LitSegger:
    """
    Load a LitSegger model from a checkpoint.

    Parameters
    ----------
    checkpoint_path : str
        Specific checkpoint file to load, or directory where the model checkpoints are stored.
        If directory, the latest checkpoint is loaded.

    Returns
    -------
    LitSegger
        The loaded LitSegger model.

    Raises
    ------
    FileNotFoundError
        If the specified checkpoint file does not exist.
    """
    checkpoint_path = Path(checkpoint_path)
    msg = f"No checkpoint found at {checkpoint_path}. Please make sure you've provided the correct path."

    # Get last checkpoint if directory is provided
    if os.path.isdir(checkpoint_path):
        checkpoints = glob.glob(str(checkpoint_path / "*.ckpt"))
        if len(checkpoints) == 0:
            raise FileNotFoundError(msg)

        # Sort checkpoints by epoch and step
        def sort_order(c):
            match = re.match(r".*epoch=(\d+)-step=(\d+).ckpt", c)
            return int(match[1]), int(match[2])

        checkpoint_path = Path(sorted(checkpoints, key=sort_order)[-1])
    elif not checkpoint_path.exists():
        raise FileExistsError(msg)

    # Load model from checkpoint
    lit_segger = LitSegger.load_from_checkpoint(
        checkpoint_path=checkpoint_path,
    )

    return lit_segger


def get_similarity_scores(
    model: torch.nn.Module,
    batch: Batch,
    from_type: str,
    to_type: str,
    receptive_field: dict,
    knn_method: str = "cuda",
    gpu_id: int = 0,  # Added argument for GPU ID
) -> coo_matrix:
    """
    Compute similarity scores between embeddings for 'from_type' and 'to_type' nodes
    using sparse matrix multiplication with CuPy and the 'sees' edge relation.

    Args:
        model (torch.nn.Module): The segmentation model used to generate embeddings.
        batch (Batch): A batch of data containing input features and edge indices.
        from_type (str): The type of node from which the similarity is computed.
        to_type (str): The type of node to which the similarity is computed.
        knn_method (str, optional): The method to use for nearest neighbors. Defaults to 'cuda'.
        gpu_id (int, optional): The GPU ID to use for the computations. Defaults to 0.

    Returns:
        coo_matrix: A sparse matrix containing the similarity scores between
                    'from_type' and 'to_type' nodes.
    """

    with cp.cuda.Device(gpu_id):
        # Move the batch to the specified GPU
        batch = batch.to(f"cuda:{gpu_id}")

        # Step 1: Get embeddings from the model (on GPU)
        shape = batch[from_type].x.shape[0], batch[to_type].x.shape[0]
        # Compute edge indices using knn method (still on GPU)
        edge_index = get_edge_index(
            batch[to_type].pos[:, :2].cpu(),  # 'tx' positions
            batch[from_type].pos[:, :2].cpu(),  # 'bd' positions
            k=receptive_field[f"k_{to_type}"],
            dist=receptive_field[f"dist_{to_type}"],
            method=knn_method,
        )

        # Convert to dense adjacency matrix (on GPU)
        edge_index = coo_to_dense_adj(edge_index.T, num_nodes=shape[0], num_nbrs=receptive_field[f"k_{to_type}"])

        with torch.no_grad():
            embeddings = model(batch.x_dict, batch.edge_index_dict)

        def sparse_multiply(embeddings, edge_index, shape) -> coo_matrix:
            m = torch.nn.ZeroPad2d((0, 0, 0, 1))  # pad bottom with zeros

            similarity = torch.bmm(
                m(embeddings[to_type])[edge_index],  # 'to' x 'from' neighbors x embed
                embeddings[from_type].unsqueeze(-1),  # 'to' x embed x 1
            )  # -> 'to' x 'from' neighbors x 1
            del embeddings
            # Sigmoid to get most similar 'to_type' neighbor
            similarity[similarity == 0] = -torch.inf  # ensure zero stays zero
            similarity = F.sigmoid(similarity)
            # Neighbor-filtered similarity scores
            # shape = batch[from_type].x.shape[0], batch[to_type].x.shape[0]
            indices = torch.argwhere(edge_index != -1).T
            indices[1] = edge_index[edge_index != -1]
            rows = cp.fromDlpack(to_dlpack(indices[0, :].to("cuda")))
            columns = cp.fromDlpack(to_dlpack(indices[1, :].to("cuda")))
            # print(rows)
            del indices
            values = similarity[edge_index != -1].flatten()
            sparse_result = coo_matrix((cp.fromDlpack(to_dlpack(values)), (rows, columns)), shape=shape)
            return sparse_result
            # Free GPU memory after computation

        # Call the sparse multiply function
        sparse_similarity = sparse_multiply(embeddings, edge_index, shape)

    return sparse_similarity


def predict_batch(
    lit_segger: torch.nn.Module,
    batch: Batch,
    score_cut: float,
    receptive_field: Dict[str, float],
    use_cc: bool = True,
    knn_method: str = "cuda",
    edge_index_save_path: Union[str, Path] = None,
    output_ddf_save_path: Union[str, Path] = None,
    gpu_id: int = 0,  # Added argument for GPU ID
):
    """
    Predict cell assignments for a batch of transcript data using a segmentation model.
    Writes both the assignments and edge_index directly into Parquet files incrementally.

    Args:
        lit_segger (torch.nn.Module): The lightning module wrapping the segmentation model.
        batch (Batch): A batch of transcript and cell data.
        score_cut (float): The threshold for assigning transcripts to cells based on similarity scores.
        receptive_field (Dict[str, float]): Dictionary defining the receptive field for transcript-cell
                                            and transcript-transcript relations.
        use_cc (bool, optional): If True, perform connected components analysis for unassigned transcripts.
                                 Defaults to True.
        knn_method (str, optional): The method to use for nearest neighbors. Defaults to 'cuda'.
        edge_index_save_path (str, optional): Path to the Parquet file where edge indices are saved incrementally.
        output_ddf_save_path (str, optional): Path to the Parquet file where transcript assignments (`output_ddf`)
                                              are saved incrementally.
        gpu_id (int, optional): The GPU ID to use for the computations. Defaults to 0.
    """

    def _get_id():
        """Generate a random Xenium-style ID."""
        return "".join(np.random.choice(list("abcdefghijklmnopqrstuvwxyz"), 8)) + "-nx"

    print(gpu_id)
    with cp.cuda.Device(gpu_id):
        # Move the batch to the specified GPU
        batch = batch.to(f"cuda:{gpu_id}")
        lit_segger.model = lit_segger.model.to(f"cuda:{gpu_id}")

        # Extract transcript IDs and initialize a dictionary for assignments
        transcript_id = batch["tx"].id.cpu().numpy().astype("str")
        assignments = {"transcript_id": transcript_id}

        if len(batch["bd"].pos) >= 10:
            while True:
                try:
                    # Step 1: Compute similarity scores between 'tx' (transcripts) and 'bd' (boundaries)
                    scores = get_similarity_scores(
                        lit_segger.model, batch, "tx", "bd", receptive_field, knn_method=knn_method, gpu_id=gpu_id
                    )
                    break
                except Exception as e:
                    print(f"This weird error: {e}.")
            torch.cuda.empty_cache()

            # Convert sparse matrix to dense format (on GPU)
            dense_scores = scores.toarray()  # Convert to dense NumPy array
            del scores  # Remove from memory
            cp.get_default_memory_pool().free_all_blocks()  # Free CuPy memory

            # Step 2: Maximize score and assign transcripts based on score threshold
            belongs = cp.max(dense_scores, axis=1)  # Max score per transcript
            assignments["score"] = cp.asnumpy(belongs)  # Move back to CPU

            mask = assignments["score"] >= score_cut  # Mask for assigned transcripts
            all_ids = batch["bd"].id  # Boundary IDs as NumPy array
            assignments["segger_cell_id"] = np.where(mask, all_ids[cp.argmax(dense_scores, axis=1).get()], None)
            # Clear memory after score processing
            del dense_scores
            cp.get_default_memory_pool().free_all_blocks()  # Free CuPy memory
            torch.cuda.empty_cache()

            assignments["bound"] = np.where(mask, 1, 0)  # Mark as 'bound' (1 if assigned, 0 if unassigned)

            # Step 3: Handle unassigned transcripts with connected components (if use_cc=True)
            if use_cc:
                scores_tx = get_similarity_scores(
                    lit_segger.model, batch, "tx", "tx", receptive_field, knn_method=knn_method, gpu_id=gpu_id
                )

                # Stay on GPU and use CuPy sparse matrices
                no_id_scores = cupyx.scipy.sparse.coo_matrix(
                    (scores_tx.data, (scores_tx.row, scores_tx.col)), shape=scores_tx.shape
                )

                # Apply threshold on GPU
                no_id_scores.data[no_id_scores.data < score_cut] = 0  # Apply threshold

                # Zero out the diagonal on GPU
                no_id_scores = zero_out_diagonal_gpu(no_id_scores)
                no_id_scores.eliminate_zeros()  # Remove zero entries to keep the matrix sparse

                # Find unassigned transcripts (those with no segger_cell_id)
                no_id = cp.where(cp.asarray(assignments["segger_cell_id"] == None))[
                    0
                ]  # Using CuPy to handle None values

                if len(no_id) > 0:  # Only compute if there are unassigned transcripts
                    # Apply score cut-off to unassigned transcripts
                    no_id_scores = subset_sparse_matrix(no_id_scores, no_id, no_id)
                    no_id_scores.data[no_id_scores.data < score_cut] = 0  # Apply threshold
                    no_id_scores.eliminate_zeros()  # Clean up zeros

                    # Find the non-zero entries in the no_id_scores to construct edge_index
                    non_zero_rows, non_zero_cols, _ = find(no_id_scores)
                    unassigned_ids = transcript_id[no_id.get()]  # Unassigned transcript IDs

                    # Construct edge index (source, target) based on non-zero connections in no_id_scores
                    source_nodes = unassigned_ids[non_zero_rows.get()]
                    target_nodes = unassigned_ids[non_zero_cols.get()]
                    if len(source_nodes) > 10:
                        # while True:
                        try:
                            # # Save edge_index using CuDF and Dask-CuDF for GPU acceleration
                            edge_index_ddf = delayed(dd.from_pandas)(
                                pd.DataFrame({"source": source_nodes, "target": target_nodes}), npartitions=1
                            )
                            # Use delayed for asynchronous disk writing of edge_index in Dask DataFrame
                            delayed_write_edge_index = delayed(edge_index_ddf.to_parquet)(
                                edge_index_save_path, append=True, ignore_divisions=True
                            )
                            delayed_write_edge_index.compute()  # Schedule writing
                            # break
                        except Exception as e:
                            print(f"resource access fault, trying again.")

            assignments = {
                "transcript_id": assignments["transcript_id"].astype("str"),
                "score": assignments["score"].astype("float32"),
                "segger_cell_id": assignments["segger_cell_id"].astype("str"),  # Ensure 'string' dtype
                "bound": assignments["bound"].astype("int8"),  # Ensure 'int64' dtype
            }
            # Step 4: Convert assignments to Dask-CuDF DataFrame for this batch
            # batch_ddf = dask_cudf.from_cudf(cudf.DataFrame(assignments), npartitions=1)
            if len(assignments["transcript_id"]) > 10:
                # while True:
                try:
                    batch_ddf = delayed(dd.from_pandas)(pd.DataFrame(assignments), npartitions=1)

                    # Save the updated `output_ddf` asynchronously using Dask delayed
                    delayed_write_output_ddf = delayed(batch_ddf.to_parquet)(
                        output_ddf_save_path, append=True, ignore_divisions=True
                    )
                    delayed_write_output_ddf.compute()  # Schedule writing
                    # Free memory after computation
                    cp.get_default_memory_pool().free_all_blocks()  # Free CuPy memory
                    torch.cuda.empty_cache()
                    # break
                except Exception as e:
                    print(f"resource access fault, trying again.")


def segment(
    model: LitSegger,
    dm: SeggerDataModule,
    save_dir: Union[str, Path],
    seg_tag: str,
    transcript_file: Union[str, Path],
    score_cut: float = 0.5,
    use_cc: bool = True,
    save_transcripts: bool = True,
    save_anndata: bool = True,
    save_cell_masks: bool = False,  # Placeholder for future implementation
    receptive_field: dict = {"k_bd": 4, "dist_bd": 10, "k_tx": 5, "dist_tx": 3},
    knn_method: str = "cuda",
    verbose: bool = False,
    gpu_ids: list = ["0"],
    cleanup: bool = True,
    **anndata_kwargs,
) -> None:
    """
    Perform segmentation using the model, save transcripts, AnnData, and cell masks as needed,
    and log the parameters used during segmentation.

    Args:
        model (LitSegger): The trained segmentation model.
        dm (SeggerDataModule): The SeggerDataModule instance for data loading.
        save_dir (Union[str, Path]): Directory to save the final segmentation results.
        seg_tag (str): Tag to include in the saved filename.
        transcript_file (Union[str, Path]): Path to the transcripts Parquet file.
        score_cut (float, optional): The threshold for assigning transcripts to cells based on
                                     similarity scores. Defaults to 0.5.
        use_cc (bool, optional): If True, perform connected components analysis for unassigned
                                 transcripts. Defaults to True.
        save_transcripts (bool, optional): Whether to save the transcripts as Parquet. Defaults to True.
        save_anndata (bool, optional): Whether to save the results in AnnData format. Defaults to True.
        save_cell_masks (bool, optional): Save cell masks as Dask Geopandas Parquet. Defaults to False.
        receptive_field (dict, optional): Defines the receptive field for transcript-cell and
                                          transcript-transcript relations. Defaults to
                                          `{'k_bd': 4, 'dist_bd': 10, 'k_tx': 5, 'dist_tx': 3}`.
        knn_method (str, optional): The method to use for nearest neighbors (`'cuda'` or `'kd_tree'`).
                                    Defaults to 'cuda'.
        verbose (bool, optional): Whether to print verbose status updates. Defaults to False.
        cleanup (bool, optional): To clean up intermediate files, i.e., `transcripts_df.parqeut` and `edge_index.parquet`. Defaults to True.
        **anndata_kwargs: Additional keyword arguments passed to the `create_anndata` function.

    Returns:
        None. Saves the result to disk in various formats and logs the parameter choices.
    """

    start_time = time()

    # Create a subdirectory with important parameter info (receptive field values)
    sub_dir_name = f"{seg_tag}_{score_cut}_{use_cc}_{receptive_field['k_bd']}_{receptive_field['dist_bd']}_{receptive_field['k_tx']}_{receptive_field['dist_tx']}_{datetime.now().strftime('%Y%m%d')}"
    save_dir = Path(save_dir) / sub_dir_name
    save_dir.mkdir(parents=True, exist_ok=True)

    # Paths for saving the output_ddf and edge_index Parquet files
    output_ddf_save_path = save_dir / "transcripts_df.parquet"
    edge_index_save_path = save_dir / "edge_index.parquet"

    if output_ddf_save_path.exists():
        warnings.warn(f"Removing existing file: {output_ddf_save_path}")
        shutil.rmtree(output_ddf_save_path)

    if use_cc:
        if edge_index_save_path.exists():
            warnings.warn(f"Removing existing file: {edge_index_save_path}")
            shutil.rmtree(edge_index_save_path)

    if verbose:
        print(f"Assigning transcripts to nuclei for {seg_tag}...")

    # Step 1: Load the data loaders from the SeggerDataModule
    step_start_time = time()
    # train_dataloader = dm.train_dataloader()
    # val_dataloader = dm.val_dataloader()
    # test_dataloader = dm.test_dataloader()

    # # Initialize Dask DataFrame for assignments
    # output_ddf = None
    cluster = LocalCUDACluster(
        CUDA_VISIBLE_DEVICES=gpu_ids,
        n_workers=len(gpu_ids),  # Number of workers to use (adjust based on your available CPUs/GPUs)
        threads_per_worker=1,  # Each worker will use 1 thread
    )

    # # Connect the client to the cluster
    client = Client(cluster)

    @dask.delayed
    def process_batch(batch, gpu_id):
        # Assume you're using CuPy, and you need to use a specific GPU
        predict_batch(
            model,
            batch,
            score_cut,
            receptive_field,
            use_cc=use_cc,
            knn_method=knn_method,
            edge_index_save_path=edge_index_save_path,
            output_ddf_save_path=output_ddf_save_path,
            gpu_id=gpu_id,
        )

    # this is to off-load some tasks from the computation graph
    for j in range(len(dm.train) // 100):
        print(len(dm.train))
        ind_min = j * 100
        ind_max = min((j + 1) * 100, len(dm.train))
        delayed_tasks = [
            process_batch(batch, gpu_ids[i % len(gpu_ids)]) for i, batch in enumerate(dm.train[ind_min:ind_max])
        ]
        future = dask.persist(*delayed_tasks)
        progress(future)
        dask.compute(*future)
        torch.cuda.empty_cache()
        cp.get_default_memory_pool().free_all_blocks()

    for j in range(len(dm.val) // 50):
        ind_min = j * 50
        ind_max = min((j + 1) * 50, len(dm.val))
        delayed_tasks = [
            process_batch(batch, gpu_ids[i % len(gpu_ids)]) for i, batch in enumerate(dm.val[ind_min:ind_max])
        ]
        future = dask.persist(*delayed_tasks)
        progress(future)
        dask.compute(*future)
        torch.cuda.empty_cache()
        cp.get_default_memory_pool().free_all_blocks()

    for j in range(len(dm.test) // 50):
        ind_min = j * 50
        ind_max = min((j + 1) * 50, len(dm.test))
        delayed_tasks = [
            process_batch(batch, gpu_ids[i % len(gpu_ids)]) for i, batch in enumerate(dm.test[ind_min:ind_max])
        ]
        future = dask.persist(*delayed_tasks)
        progress(future)
        dask.compute(*future)
        torch.cuda.empty_cache()
        cp.get_default_memory_pool().free_all_blocks()

    client.close()

    # pqdm(delayed_tasks, n_jobs=len(gpu_ids), argument_type='delayed', progress_bar=True)
    # dask.compute(*delayed_tasks)
    # delayed_tasks = [process_batch(batch, gpu_ids[i % len(gpu_ids)]) for i, batch in enumerate(batches)]

    # # Use tqdm for progress bar
    # with ProgressBar():
    #     # Execute the delayed tasks with a Dask compute call
    #     dask.compute(*delayed_tasks)

    # Loop through the data loaders (train, val, and test)
    # for loader_name, loader in zip(
    #     ["Train", "Validation", "Test"], [train_dataloader, val_dataloader, test_dataloader]
    # ):
    #     # for loader_name, loader in zip(['Test'], [test_dataloader]):
    #     if verbose:
    #         print(f"Processing {loader_name} data...")

    #     for batch in tqdm(loader, desc=f'Processing {loader_name} batches'):
    #         gpu_id = random.choice(gpu_ids)
    #         # Call predict_batch for each batch
    #         predict_batch(
    #             model,
    #             batch,
    #             score_cut,
    #             receptive_field,
    #             use_cc=use_cc,
    #             knn_method=knn_method,
    #             edge_index_save_path=edge_index_save_path,
    #             output_ddf_save_path=output_ddf_save_path,
    #             gpu_id=gpu_id
    #         )

    # if verbose:
    #     elapsed_time = time() - step_start_time
    #     print(f"Batch processing completed in {elapsed_time:.2f} seconds.")
    if verbose:
        elapsed_time = time() - step_start_time
        print(f"Assigned transcripts to nuclei in {elapsed_time:.2f} seconds.")
    # Load the full saved segmentation results
    seg_final_dd = dd.read_parquet(output_ddf_save_path, ignore_metadata_file=True)

    step_start_time = time()
    if verbose:
        print(f"Applying max score selection logic...")

    # Step 1: Find max bound indices (bound == 1) and max unbound indices (bound == 0)
    max_bound_idx = seg_final_dd[seg_final_dd["bound"] == 1].groupby("transcript_id")["score"].idxmax()
    max_unbound_idx = seg_final_dd[seg_final_dd["bound"] == 0].groupby("transcript_id")["score"].idxmax()

    # Step 2: Combine indices, prioritizing bound=1 scores
    final_idx = max_bound_idx.compute().combine_first(max_unbound_idx.compute())

    # Step 3: Use the computed final_idx to select the best assignments
    # Make sure you are using the divisions and set the index correctly before loc
    # seg_final_dd = seg_final_dd.set_index('transcript_id', sorted=True)
    seg_final_filtered = seg_final_dd.loc[final_idx].compute()

    if verbose:
        elapsed_time = time() - step_start_time
        print(f"Max score selection completed in {elapsed_time:.2f} seconds.")

    # Step 3: Load the transcripts DataFrame and merge results

    if verbose:
        print(f"Loading transcripts from {transcript_file}...")

    transcripts_df = dd.read_parquet(transcript_file)
    transcripts_df["transcript_id"] = transcripts_df["transcript_id"].astype(str)

    step_start_time = time()
    if verbose:
        print(f"Merging segmentation results with transcripts...")

    # Outer merge to include all transcripts, even those without assigned cell ids
    transcripts_df_filtered = transcripts_df.merge(seg_final_filtered, on="transcript_id", how="outer")

    if verbose:
        elapsed_time = time() - step_start_time
        print(f"Merging segmentation results with transcripts completed in {elapsed_time:.2f} seconds.")

    # Step 4: Handle unassigned transcripts using connected components (if use_cc=True)
    if use_cc:

        step_start_time = time()
        if verbose:
            print(f"Regrouping the unassigned transcripts...")
        # Load edge indices from saved Parquet
        edge_index_dd = dd.read_parquet(edge_index_save_path)

        # Step 2: Get unique transcript_ids from edge_index_dd and their positional indices
        transcript_ids_in_edges = dd.concat([edge_index_dd["source"], edge_index_dd["target"]]).unique().compute()

        # Create a lookup table with unique indices
        lookup_table = pd.Series(data=range(len(transcript_ids_in_edges)), index=transcript_ids_in_edges).to_dict()

        # Map source and target to positional indices
        edge_index_dd["index_source"] = edge_index_dd["source"].map(lookup_table)
        edge_index_dd["index_target"] = edge_index_dd["target"].map(lookup_table)
        # Step 3: Compute connected components for transcripts involved in edges
        source_indices = np.asarray(edge_index_dd["index_source"].compute())
        target_indices = np.asarray(edge_index_dd["index_target"].compute())
        data_cp = np.ones(len(source_indices), dtype=cp.float32)

        # Create the sparse COO matrix
        coo_cp_matrix = scipy_coo_matrix(
            (data_cp, (source_indices, target_indices)),
            shape=(len(transcript_ids_in_edges), len(transcript_ids_in_edges)),
        )

        # Use CuPy's connected components algorithm to compute components
        n, comps = cc(coo_cp_matrix, directed=True, connection="weak")

        # Step 4: Map back the component labels to the original transcript_ids
        comp_labels = pd.Series(comps, index=transcript_ids_in_edges)
        # Step 5: Handle only unassigned transcripts in transcripts_df_filtered
        unassigned_mask = transcripts_df_filtered["segger_cell_id"].isna()

        unassigned_transcripts_df = transcripts_df_filtered.loc[unassigned_mask, ["transcript_id"]]

        # Step 6: Map component labels only to unassigned transcript_ids
        new_segger_cell_ids = unassigned_transcripts_df["transcript_id"].map(comp_labels)

        # Step 7: Create a DataFrame with updated 'segger_cell_id' for unassigned transcripts
        unassigned_transcripts_df = unassigned_transcripts_df.assign(segger_cell_id=new_segger_cell_ids)

        # Step 8: Merge this DataFrame back into the original to update only the unassigned segger_cell_id
        # We perform a left join so that only the rows in unassigned_transcripts_df are updated
        # transcripts_df_filtered = transcripts_df_filtered.drop(columns='segger_cell_id')

        # Merging the updates back to the original DataFrame
        transcripts_df_filtered = transcripts_df_filtered.merge(
            unassigned_transcripts_df[["transcript_id", "segger_cell_id"]],
            on="transcript_id",
            how="left",  # Perform a left join to only update the unassigned rows
            suffixes=("", "_new"),  # Suffix for new column to avoid overwriting
        )

        # Step 9: Fill missing segger_cell_id values with the updated values from the merge
        transcripts_df_filtered["segger_cell_id"] = transcripts_df_filtered["segger_cell_id"].fillna(
            transcripts_df_filtered["segger_cell_id_new"]
        )

        # Step 10: Clean up by dropping the temporary 'segger_cell_id_new' column
        transcripts_df_filtered = transcripts_df_filtered.drop(columns=["segger_cell_id_new"])

        # Fill the NaN values in segger_cell_id with the already existing (assigned) values
        # transcripts_df_filtered['segger_cell_id'] = transcripts_df_filtered['segger_cell_id'].fillna(transcripts_df_filtered['segger_cell_id_target'])

        # Drop any temporary columns used during the merge
        # transcripts_df_filtered = transcripts_df_filtered.drop(columns=['segger_cell_id_target'])

        if verbose:
            elapsed_time = time() - step_start_time
            print(f"Regrouped the unassigned transcripts in {elapsed_time:.2f} seconds.")

    # Step 5: Save the merged results based on options

    if save_transcripts:
        if verbose:
            step_start_time = time()
            print(f"Saving transcripts.parquet...")
        transcripts_save_path = save_dir / "segger_transcripts.parquet"
        # transcripts_df_filtered = transcripts_df_filtered.repartition(npartitions=100)
        transcripts_df_filtered.to_parquet(
            transcripts_save_path,
            engine="pyarrow",  # PyArrow is faster and recommended
            compression="snappy",  # Use snappy compression for speed
            write_index=False,  # Skip writing index if not needed
            append=False,  # Set to True if you're appending to an existing Parquet file
            overwrite=True,
        )  # Dask handles Parquet well
        if verbose:
            elapsed_time = time() - step_start_time
            print(f"Saved trasncripts.parquet in {elapsed_time:.2f} seconds.")

    if save_anndata:
        if verbose:
            step_start_time = time()
            print(f"Saving anndata object...")
        anndata_save_path = save_dir / "segger_adata.h5ad"
        segger_adata = create_anndata(transcripts_df_filtered.compute(), **anndata_kwargs)  # Compute for AnnData
        segger_adata.write(anndata_save_path)
        if verbose:
            elapsed_time = time() - step_start_time
            print(f"Saved anndata object in {elapsed_time:.2f} seconds.")

    if save_cell_masks:
        if verbose:
            step_start_time = time()
            print(f"Computing and saving cell masks anndata object...")
        # Placeholder for future cell masks implementation as Dask Geopandas Parquet
        cell_masks_save_path = save_dir / "segger_cell_boundaries.parquet"
        if verbose:
            elapsed_time = time() - step_start_time
            print(f"Saved cell masks in {elapsed_time:.2f} seconds.")

    if cleanup:
        if verbose:
            step_start_time = time()
            print(f"Cleaning up intermediate files...")
        shutil.rmtree(output_ddf_save_path)
        shutil.rmtree(edge_index_save_path)

    # Step 6: Save segmentation parameters as a JSON log
    log_data = {
        "seg_tag": seg_tag,
        "score_cut": score_cut,
        "use_cc": use_cc,
        "receptive_field": receptive_field,
        "knn_method": knn_method,
        "save_transcripts": save_transcripts,
        "save_anndata": save_anndata,
        "save_cell_masks": save_cell_masks,
        "timestamp": datetime.now().isoformat(),
    }

    log_path = save_dir / "segmentation_log.json"
    with open(log_path, "w") as log_file:
        json.dump(log_data, log_file, indent=4)

    # Step 7: Garbage collection and memory cleanup
    torch.cuda.empty_cache()
    gc.collect()

    # Total time taken for the segmentation process
    if verbose:
        total_time = time() - start_time
        print(f"Total segmentation process completed in {total_time:.2f} seconds.")



import click
import yaml
from pydoc import locate
from argparse import Namespace
import typing
import os
import logging


def add_options(
    config_path: os.PathLike,
    show_default: bool = True,
):
    """
    A decorator to add command-line options to a Click command from a YAML
    configuration file.

    Parameters:
    config_path (os.PathLike): The path to the YAML configuration file
    containing the options.
    show_default (bool): Whether to show default values in help.

    Returns:
    function: A decorator function that adds the options to the Click command.

    The YAML configuration file should have the following format:
    ```
    option_name:
        type: "type_name"  # Optional, the type of the option
        (e.g., "str", "int")
        help: "description"  # Optional, the help text for the option
        default: value  # Optional, the default value for the option
        required: true/false  # Optional, whether the option is required
        ...
    ```
    Example usage:
    ```
    # config.yaml
    name:
        type: "str"
        help: "Your name"
        required: true
    age:
        type: "int"
        help: "Your age"
        default: 30

    # script.py
    @add_options('config.yaml')
    @click.command()
    def greet(args):
        click.echo(f"Hello, {args.name}! You are {args.age} years old.")
    ```
    """

    def decorator(function: typing.Callable):
        # Wrap the original function to convert kwargs to a Namespace object
        def wrapper(**kwargs):
            args_namespace = Namespace(**kwargs)
            return function(args_namespace)

        # Load the YAML configuration file
        with open(config_path, "r") as file:
            config = yaml.safe_load(file.read())

        # Decorate function with all options
        for name, kwargs in reversed(config.items()):
            kwargs["show_default"] = show_default
            if "type" in kwargs:
                kwargs["type"] = locate(kwargs["type"])
            wrapper = click.option(f"--{name}", **kwargs)(wrapper)

        return wrapper

    return decorator


class CustomFormatter(logging.Formatter):
    """
    Custom logging formatter to add color-coded log levels to the log messages.

    Attributes:
    grey (str): ANSI escape code for grey color.
    yellow (str): ANSI escape code for yellow color.
    red (str): ANSI escape code for red color.
    bold_red (str): ANSI escape code for bold red color.
    reset (str): ANSI escape code to reset color.
    format (str): The format string for log messages.
    FORMATS (dict): A dictionary mapping log levels to their respective
    color-coded format strings.

    Methods:
    format(record):
        Format the specified record as text, applying color codes based on the
        log level.
    """

    grey = "\x1b[38;20m"
    green = "\x1b[32;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s %(levelname)s: %(message)s"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: green + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



from __future__ import annotations

from collections import defaultdict
from typing import Iterable, Tuple

import numpy as np
import torch


def _grid_offsets(connectivity: int) -> Tuple[Tuple[int, int], ...]:
    if connectivity == 4:
        return ((-1, 0), (1, 0), (0, -1), (0, 1))
    if connectivity == 8:
        return (
            (-1, 0),
            (1, 0),
            (0, -1),
            (0, 1),
            (-1, -1),
            (-1, 1),
            (1, -1),
            (1, 1),
        )
    msg = f"Unsupported grid connectivity: {connectivity}. Expected 4 or 8."
    raise ValueError(msg)


def build_grid_bin_edge_index(
    bx: Iterable[int] | np.ndarray | torch.Tensor,
    by: Iterable[int] | np.ndarray | torch.Tensor,
    connectivity: int = 8,
) -> Tuple[torch.Tensor, np.ndarray]:
    bx = np.asarray(bx)
    by = np.asarray(by)
    if bx.size == 0:
        return torch.empty((2, 0), dtype=torch.long), np.empty((0, 2), dtype=int)

    bins = np.unique(np.stack([bx, by], axis=1), axis=0)
    bin_to_idx = {(int(x), int(y)): i for i, (x, y) in enumerate(bins)}

    offsets = _grid_offsets(connectivity)
    src = []
    dst = []
    for i, (x, y) in enumerate(bins):
        x = int(x)
        y = int(y)
        for dx, dy in offsets:
            j = bin_to_idx.get((x + dx, y + dy))
            if j is not None and j != i:
                src.append(i)
                dst.append(j)

    if len(src) == 0:
        return torch.empty((2, 0), dtype=torch.long), bins

    edge_index = torch.tensor([src, dst], dtype=torch.long).contiguous()
    return edge_index, bins


def build_grid_gene_bin_edge_index(
    bx: Iterable[int] | np.ndarray | torch.Tensor,
    by: Iterable[int] | np.ndarray | torch.Tensor,
    connectivity: int = 8,
    within_bin_edges: str = "none",
) -> torch.Tensor:
    """
    Builds edge index for grid-based gene-bin nodes.
    Each node corresponds to a specific (gene_id, bx, by).
    
    Nodes within the same bin are connected via 'within_bin_edges' strategy.
    Nodes in adjacent bins are connected via 'connectivity' strategy (connecting hubs).
    """
    bx = np.asarray(bx)
    by = np.asarray(by)
    n_nodes = bx.shape[0]

    if n_nodes == 0:
        return torch.empty((2, 0), dtype=torch.long)

    # Group nodes by bin
    bin_nodes = defaultdict(list)
    for i, (x, y) in enumerate(zip(bx, by)):
        bin_nodes[(int(x), int(y))].append(i)

    src = []
    dst = []

    # 1. Within-bin edges (Star topology)
    # The first node in the list is the "hub".
    hubs = {}  # Map (x, y) -> hub_node_idx
    
    if within_bin_edges == "star":
        for (x, y), nodes in bin_nodes.items():
            if not nodes:
                continue
            hub = nodes[0]
            hubs[(x, y)] = hub
            
            # Connect all other nodes to the hub (undirected)
            for other in nodes[1:]:
                src.extend([hub, other])
                dst.extend([other, hub])
    elif within_bin_edges == "none":
        # Even if "none", we need to identify a hub to connect to neighbors
        for (x, y), nodes in bin_nodes.items():
            if nodes:
                hubs[(x, y)] = nodes[0]
    else:
        raise ValueError(f"Unsupported within_bin_edges: {within_bin_edges}")

    # 2. Between-bin edges (Grid topology between hubs)
    offsets = _grid_offsets(connectivity)
    
    for (x, y), hub_idx in hubs.items():
        for dx, dy in offsets:
            neighbor_hub = hubs.get((x + dx, y + dy))
            if neighbor_hub is not None:
                src.append(hub_idx)
                dst.append(neighbor_hub)

    if len(src) == 0:
        return torch.empty((2, 0), dtype=torch.long)

    edge_index = torch.tensor([src, dst], dtype=torch.long).contiguous()
    return edge_index


import unittest
import torch
from segger.models.segger_model import Segger
from torch_geometric.nn import to_hetero
from torch_geometric.data import HeteroData


class TestSeggerModel(unittest.TestCase):

    def setUp(self):
        model = Segger(init_emb=16, hidden_channels=32, out_channels=32, heads=3)
        metadata = (["tx", "nc"], [("tx", "belongs", "nc"), ("tx", "neighbors", "tx")])
        self.model = to_hetero(model, metadata=metadata, aggr="sum")
        self.data = HeteroData()
        self.data["tx"].x = torch.randn(10, 16)
        self.data["nc"].x = torch.randn(5, 16)
        self.data["tx", "belongs", "nc"].edge_index = torch.tensor([[0, 1, 2], [0, 1, 2]], dtype=torch.long)
        self.data["tx", "neighbors", "tx"].edge_index = torch.tensor([[0, 1], [1, 2]], dtype=torch.long)

    def test_forward(self):
        out = self.model(self.data.x_dict, self.data.edge_index_dict)
        self.assertTrue("tx" in out)
        self.assertTrue("nc" in out)
        self.assertEqual(out["tx"].shape[1], 32 * 3)
        self.assertEqual(out["nc"].shape[1], 32 * 3)

    """
    def test_decode(self):
        z = {'tx': torch.randn(10, 16), 'nc': torch.randn(5, 16)}
        edge_label_index = torch.tensor([[0, 1, 2], [0, 1, 2]], dtype=torch.long)
        out = self.model.decode(z, edge_label_index)
        self.assertEqual(out.shape[0], 3)
    """


if __name__ == "__main__":
    unittest.main()
